"""Functionality for removing backups from a repo's history"""
import datetime as dt
import logging
from pathlib import Path

from . import _git, backup
from .logging import IMPORTANT

LOGGER = logging.getLogger(__name__)


def rewrite_history(repo_root: Path, starting_point: str, *revisions: str) -> str:
    """Rewrite the repo's history by only including the specified backups,
    effectively deleting the ones in between

    Parameters
    ----------
    repo_root : Path
        The directory containing the GSB-managed repo
    starting_point: str
        The commit hash or tag name to start revising from (all prior backups
        will be kept)
    *revisions: str
        The commit hashes / tag names for the backups that should be included
        / kept in the new history

    Returns
    -------
    str
        The tag name or commit hash for the most recent backup in the rewritten
        history

    Notes
    -----
    - The current repo state will always be kept (and, in the case that there
      are un-backed-up changes, those changes will be backed up before the
      history is rewritten).
    - The ordering of the provided revisions is not checked in advance, nor is
      anything done to check for duplicates. Providing the backups out-of-order
      will create a new history that frames the backups in the order provided.

    Raises
    ------
    OSError
        If the specified repo does not exist or is not a gsb-managed repo
    ValueError
        If any of the specified revisions do not exist
    """
    _ = _git.show(repo_root, starting_point)  # ensure starting point is valid
    tag_lookup = {
        tag.target: tag for tag in _git.get_tags(repo_root, annotated_only=False)
    }
    new_history: list[_git.Tag | _git.Commit] = []

    for reference in revisions:
        revision: _git.Tag | _git.Commit = _git.show(repo_root, reference)
        if revision in tag_lookup.keys():
            revision = tag_lookup[revision]  # type: ignore[index]
        new_history.append(revision)

    try:
        head = backup.create_backup(repo_root)
        LOGGER.log(IMPORTANT, "Unsaved changes have been backed up as %s", head[:8])
        new_history.append(_git.show(repo_root, head))
    except ValueError:
        # nothing to back up
        pass

    try:
        branch_name = dt.datetime.now().strftime("gsb_rebase_%Y.%m.%d+%H%M%S")
        _git.checkout_branch(repo_root, branch_name, starting_point)
        head = starting_point
        tags_to_update: list[tuple[_git.Tag, str]] = []
        for revision in new_history:
            match revision:
                case _git.Commit():
                    _git.reset(repo_root, revision.hash, hard=True)
                    _git.reset(repo_root, head, hard=False)
                    new_hash = _git.commit(
                        repo_root,
                        message=(
                            revision.message + "\n\n" + f"rebase of {revision.hash}"
                        ),
                        timestamp=revision.timestamp,
                    ).hash
                    head = new_hash
                case _git.Tag():
                    _git.reset(repo_root, revision.target.hash, hard=True)
                    _git.reset(repo_root, head, hard=False)
                    new_hash = _git.commit(
                        repo_root,
                        message=(
                            (revision.annotation or revision.name)
                            + "\n\n"
                            + f"rebase of {revision.target.hash}"
                            + f' ("{revision.target.message.strip()}")'
                        ),
                        timestamp=revision.target.timestamp,
                    ).hash
                    tags_to_update.append((revision, new_hash))
                    head = new_hash
                case _:  # pragma: no cover
                    raise NotImplementedError(
                        f"Don't know how to handle revision of type {type(revision)}"
                    )
        for tag, target in tags_to_update:
            _git.delete_tag(repo_root, tag.name)
            _git.tag(repo_root, tag.name, tag.annotation, target=target)
        try:
            _git.delete_branch(repo_root, "gsb")
        except ValueError as delete_fail:
            # this can happen if you onboarded an existing repo to gsb, in
            # which case the active branch won't necessarily be gsb
            LOGGER.warning("Could not delete branch %s:\n    %s", "gsb", delete_fail)
        _git.checkout_branch(repo_root, "gsb", head)
        return head
    except Exception as something_went_wrong:  # pragma: no cover
        _git.reset(repo_root, head, hard=True)
        raise something_went_wrong
    finally:
        _git.checkout_branch(repo_root, "gsb", None)
        _git.delete_branch(repo_root, branch_name)


def delete_backups(repo_root: Path, *revisions: str) -> str:
    """Delete the specified backups

    Parameters
    ----------
    repo_root : Path
        The directory containing the GSB-managed repo
    *revisions : str
        The commit hashes and tag names of the backups to delete

    Returns
    -------
    str
        The tag name or commit hash for the most recent backup in the rewritten
        history

    Notes
    -----
    - The current repo state will always be kept (and, in the case that there
      are un-backed-up changes, those changes will be backed up before the
      history is rewritten).
    - Deleting the initial backup is not currently supported.

    Raises
    ------
    OSError
        If the specified repo does not exist or is not a gsb-managed repo
    ValueError
        If the specified revision does not exist
    """
    to_delete: dict[_git.Commit, str] = {}
    for revision in revisions:
        match reference := _git.show(repo_root, revision):
            case _git.Commit():
                # keep the link back to the revision in case we need to raise
                # an error about it later
                to_delete[reference] = revision
            case _git.Tag():
                to_delete[reference.target] = revision
            case _:  # pragma: no cover
                raise NotImplementedError(f"Don't know how to handle {type(reference)}")

    to_keep: list[str] = []
    for commit in _git.log(repo_root):
        if commit in to_delete:
            del to_delete[commit]
        else:
            to_keep.insert(0, commit.hash)
            if len(to_delete) == 0:
                break
    else:
        if len(to_delete) == 0:
            raise NotImplementedError(
                "Deleting the initial backup is not currently supported."
            )
        raise ValueError(
            "The following revisions exist, but they are not within"
            " the linear commit history:\n"
            + "\n".join((f"  - {revision}" for revision in to_delete.values()))
        )
    return rewrite_history(repo_root, *to_keep)



"""Functionality for tracking and managing revision history"""
import datetime as dt
import logging
from pathlib import Path
from typing import Any, TypedDict

from . import _git
from .logging import IMPORTANT

LOGGER = logging.getLogger(__name__)


class Revision(TypedDict):
    """Metadata on a GSB-managed version

    Parameters
    ----------
    identifier : str
        A short, unique identifier for the revision
    commit_hash : str
        The full hexadecimal hash associated with the revision
    description : str
        A description of the version
    timestamp : dt.datetime
        The time at which the version was created
    tagged : bool
        Whether or not this is a tagged revision
    gsb : bool
        Whether or not this is a GSB-created revision
    """

    identifier: str
    commit_hash: str
    description: str
    timestamp: dt.datetime
    tagged: bool
    gsb: bool


def get_history(
    repo_root: Path,
    tagged_only: bool = True,
    include_non_gsb: bool = False,
    limit: int = -1,
    since: dt.date = dt.datetime(1970, 1, 1),
    since_last_tagged_backup: bool = False,
    always_include_latest: bool = False,
) -> list[Revision]:
    """Retrieve a list of GSB-managed versions

    Parameters
    ----------
    repo_root : Path
        The directory containing the GSB-managed repo
    tagged_only : bool, optional
        By default, this method only returns tagged backups. To include
        all available revisions, pass in `tagged_only=False`.
    include_non_gsb : bool, optional
        By default, this method excludes any revisions created outside of `gsb`.
        To include all git commits and tags, pass in `include_non_gsb=True`.
    limit : int, optional
        By default, this method returns the entire history. To return only the
        last _n_ revisions, pass in `limit=n`.
    since : date or timestamp, optional
        By default, this method returns the entire history. To return only
        revisions made on or after a certain date, pass in `since=<start_date>`.
    since_last_tagged_backup: bool, optional
        False by default. To return only revisions made since the last tagged
        backup, pass in `since_last_tagged_backup=True` (and, presumably,
        `tagged_only=False`). This flag is compatible with all other filters.
    always_include_latest: bool, optional
        Whether to always include the latest backup, whether or not it's
        tagged or GSB-managed, and ignoring any other options. Default is False.

    Returns
    -------
    list of dict
        metadata on the requested revisions, sorted in reverse-chronological
        order

    Raises
    ------
    OSError
        If the specified repo does not exist or is not a git repo
    ValueError
        If called with `return_parent=True` `get_history` and the earliest commit
        had no parent (being the initial commit, itself).
    """
    tag_lookup = {
        tag.target: tag for tag in _git.get_tags(repo_root, annotated_only=True)
    }
    LOGGER.debug("Retrieved %s tags", len(tag_lookup))

    revisions: list[Revision] = []
    defer_break = False
    for commit in _git.log(repo_root):
        if len(revisions) == limit or commit.timestamp < since:
            if always_include_latest:
                defer_break = True
            else:
                break
        if tag := tag_lookup.get(commit):
            if since_last_tagged_backup:
                break
            tagged = True
            identifier = tag.name
            is_gsb = tag.gsb if tag.gsb is not None else commit.gsb
            description = tag.annotation or commit.message
        else:
            if tagged_only and not always_include_latest:
                continue
            tagged = False
            identifier = commit.hash[:8]
            is_gsb = commit.gsb
            description = commit.message
        if not include_non_gsb and not is_gsb and not always_include_latest:
            continue
        revisions.append(
            {
                "identifier": identifier,
                "commit_hash": commit.hash,
                "description": description.strip(),
                "timestamp": commit.timestamp,
                "tagged": tagged,
                "gsb": is_gsb,
            }
        )
        if always_include_latest:
            if defer_break:
                break
            always_include_latest = False
    return revisions


def log_revision(revision: Revision, idx: int | None) -> None:
    """Print (log) a revision

    Parameters
    ----------
    revision : dict
        Metadata for the revision
    idx : int | None
        The index to give to the revision. If None is specified, the revision
        will be displayed with a "-" instead of a numbering.

    Notes
    -----
    - The version identifiers and dates are logged at the IMPORTANT (verbose=0) level
    - The version descriptions are logged at the INFO (verbose=1) level
    - The full version hashes are logged at the DEBUG (verbose=2) level
    """
    args: list[Any] = [revision["identifier"], revision["timestamp"].isoformat("-")]
    if idx is None:
        format_string = "- %s from %s"
    else:
        format_string = "%d. %s from %s"
        args.insert(0, idx)

    LOGGER.log(IMPORTANT, format_string, *args)

    LOGGER.debug("Full reference: %s", revision["commit_hash"])
    LOGGER.info("%s", revision["description"])


def show_history(
    repo_root: Path,
    numbering: int | None = 1,
    **kwargs,
) -> list[Revision]:
    """Fetch and print (log) the list of versions for the specified repo matching
    the given specs

    Parameters
    ----------
    repo_root : Path
        The directory containing the GSB-managed repo
    numbering: int or None, optional
        When displaying the versions, the default behavior is to number the
        results, starting at 1. To set a different starting number, provide that.
        To use "-" instead of numbers, pass in `numbering=None`.
    **kwargs
        Any other options will be passed directly to `get_history()`
        method

    Notes
    -----
    See `log_revision()` for details about what information is logged to each
    log level

    Returns
    -------
    list of dict
        metadata on the requested revisions, sorted in reverse-chronological
        order

    Raises
    ------
    OSError
        If the specified repo does not exist or is not a git repo
    """
    history = get_history(repo_root, **kwargs)
    for i, revision in enumerate(history):
        log_revision(revision, i + numbering if numbering is not None else None)
    return history



"""Tests for creating new repos"""
import os
import subprocess
from pathlib import Path

import pytest

from gsb import _git, onboard
from gsb.manifest import MANIFEST_NAME, Manifest


class TestFreshInit:
    def test_root_must_be_a_directory(self, tmp_path):
        not_a_dir = tmp_path / "file.txt"
        not_a_dir.write_text("I'm a file\n")

        with pytest.raises(NotADirectoryError):
            _ = onboard.create_repo(not_a_dir)

    def test_root_must_exist(self, tmp_path):
        does_not_exist = tmp_path / "phantom"

        with pytest.raises(FileNotFoundError):
            _ = onboard.create_repo(does_not_exist)

    @pytest.fixture
    def root(self, tmp_path):
        root = tmp_path / "rootabaga"
        root.mkdir()
        yield root

    def test_no_pattern_means_add_all(self, root):
        manifest = onboard.create_repo(root)
        assert manifest.patterns == (".",)

    def test_default_behavior_gets_name_from_directory(self, root):
        assert onboard.create_repo(root).name == "rootabaga"

    def test_providing_patterns(self, root):
        manifest = onboard.create_repo(root, "savey_mcsavegame", "logs/")
        assert manifest.patterns == tuple(
            sorted(
                (
                    "savey_mcsavegame",
                    "logs/",
                )
            )
        )

    def test_init_always_creates_a_gitignore(self, root):
        _ = onboard.create_repo(root)
        _ = (root / ".gitignore").read_text()

    def test_providing_ignore(self, root):
        _ = onboard.create_repo(root, "savey_mcsavegame", ignore=[".stuff"])
        ignored = (root / ".gitignore").read_text().splitlines()
        assert ".stuff" in ignored

    def test_repo_must_not_already_exist(self, root):
        _ = onboard.create_repo(root)

        with pytest.raises(FileExistsError):
            _ = onboard.create_repo(root)

    def test_init_adds_save_contents(self, root):
        (root / "game.sav").write_text("poke\n")
        _ = onboard.create_repo(root, "game.sav")
        index = _git.ls_files(root)
        expected = {root / "game.sav", root / MANIFEST_NAME, root / ".gitignore"}
        assert expected == expected.intersection(index)

    def test_pattern_with_no_matches_does_not_error(self, root):
        (root / "game.sav").write_text("poke\n")
        _ = onboard.create_repo(root, "save.game")
        index = _git.ls_files(root)
        expected = {root / MANIFEST_NAME, root / ".gitignore"}
        assert expected == expected.intersection(index)

    def test_initial_add_respects_gitignore(self, root):
        (root / ".dot_dot").write_text("dash\n")
        _ = onboard.create_repo(root, ignore=[".*"])
        index = _git.ls_files(root)

        assert (root / ".dot_dot") not in index

    @pytest.mark.parametrize("pattern", (".dot*", "."))
    def test_gitignore_takes_priority_over_patterns_gitignore(self, root, pattern):
        (root / ".dot_dot").write_text("dash\n")
        _ = onboard.create_repo(root, pattern, ignore=[".*"])
        index = _git.ls_files(root)

        assert (root / ".dot_dot") not in index

    def test_initial_add_always_tracks_manifest_and_gitignore(self, root):
        _ = onboard.create_repo(root, ignore=[".*"])
        index = _git.ls_files(root)

        expected = {root / MANIFEST_NAME, root / ".gitignore"}
        assert expected == expected.intersection(index)

    def test_init_performs_initial_commit(self, root):
        _ = onboard.create_repo(root)
        history = _git.log(root)

        assert [commit.message for commit in history] == ["Start of gsb tracking\n"]

    def test_init_tags_that_initial_commit(self, root):
        _ = onboard.create_repo(root)
        tags = _git.get_tags(root, annotated_only=False)

        assert [tag.annotation for tag in tags] == ["Start of gsb tracking\n"]

    def test_branch_name_is_gsb_for_fresh_repo(self, root):
        _ = onboard.create_repo(root)
        assert _git._repo(root).head.shorthand == "gsb"


class TestInitExistingGitRepo:
    @pytest.fixture
    def existing_repo(self, tmp_path):
        root = tmp_path / "roto-rooter"
        root.mkdir()
        _git._repo(root, new=True, initial_branch="main")
        (root / ".gitignore").write_text(
            """# cruft
cruft

# ides
.idea
.borland_turbo
"""
        )
        yield root

    def test_init_is_fine_onboarding_an_existing_git_repo(self, existing_repo):
        _ = onboard.create_repo(existing_repo)

    def test_init_only_appends_to_existing_gitignore(self, existing_repo):
        _ = onboard.create_repo(existing_repo, ignore=["cruft", "stuff"])
        assert (
            (existing_repo / ".gitignore").read_text()
            == """# cruft
cruft

# ides
.idea
.borland_turbo

# gsb
stuff
"""
        )

    @pytest.fixture
    def repo_with_history(self, existing_repo):
        (existing_repo / "game.sav").write_text("chose a squirtle\n")
        _git.add(existing_repo, ["game.sav"])
        _git.commit(existing_repo, "Initial commit")
        _git.tag(existing_repo, "v0.0.1", "F1rst")

        (existing_repo / "game.sav").write_text("take that brock\n")
        _git.add(existing_repo, ["game.sav"])
        _git.commit(existing_repo, "Checkpoint")
        _git.tag(existing_repo, "v0.0.1+1", None)

        yield existing_repo

    def test_init_preserves_existing_commits(self, repo_with_history):
        _ = onboard.create_repo(repo_with_history)
        history = _git.log(repo_with_history)

        assert [commit.message for commit in history] == [
            "Start of gsb tracking\n",
            "Checkpoint\n",
            "Initial commit\n",
        ]

    def test_init_does_not_change_branch_name(self, repo_with_history):
        _ = onboard.create_repo(repo_with_history)
        assert _git._repo(repo_with_history).head.shorthand == "main"

    @pytest.mark.parametrize(
        "include_lightweight", (True, False), ids=("all", "annotated")
    )
    def test_init_preserves_existing_tags(self, repo_with_history, include_lightweight):
        _ = onboard.create_repo(repo_with_history)
        tags = _git.get_tags(repo_with_history, annotated_only=not include_lightweight)

        expected = {"F1rst\n", "Start of gsb tracking\n"}
        if include_lightweight:
            expected.add(None)

        assert {tag.annotation for tag in tags} == expected


class TestBackwardsCompatibility:
    def test_manifests_without_name_inherit_name_from_folder(self, tmp_path):
        (tmp_path / MANIFEST_NAME).write_text(
            """
generated_by_gsb = "0.0.2-rc2"
last_modified = "2023-08-05"
patterns = [
]
"""
        )
        assert Manifest.of(tmp_path).name == tmp_path.name


class TestCLI:
    @pytest.fixture
    def root(self, tmp_path):
        root = tmp_path / "froot"
        root.mkdir()
        (root / "toot").touch()
        yield root

    def test_init_in_cwd_by_default(self, root):
        subprocess.run(["gsb", "init"], cwd=root)
        assert (root / ".git").exists()

    def test_passing_in_a_custom_root_by_option(self, root):
        subprocess.run(["gsb", "init", "--path", root.name], cwd=root.parent)
        assert (root / ".git").exists()

    @pytest.mark.parametrize("where_is_root", ("absolute", "subdir", "pardir", "cwd"))
    def test_name_resolves_successfully(self, root, where_is_root) -> None:
        (root / "subdir").mkdir()
        if where_is_root == "absolute":
            cwd: Path | None = None
            root_arg: str = os.path.abspath(root)
        elif where_is_root == "subdir":
            cwd = root.parent
            root_arg = "froot"
        elif where_is_root == "pardir":
            cwd = root / "subdir"
            root_arg = ".."
        else:  # if where_is_root == "cwd"
            cwd = root
            root_arg = ""
        subprocess.run(["gsb", "init", "--path", root_arg], cwd=cwd)
        assert Manifest.of(root).name == "froot"

    @pytest.mark.parametrize(
        "how", ("by_argument", "by_option", "by_option_individually", "mixed")
    )
    def test_passing_in_track_patterns(self, root, how):
        patterns = {"toot", "soot", "foot", "*oot/**"}
        args = ["gsb", "init", *patterns]
        if how.startswith("by_option"):
            args.insert(2, "--track")
        if how == "by_option_individually":
            for i in range(1, len(patterns)):
                args.insert(2 + 2 * i, "--track")
        if how == "mixed":
            args.insert(2 + len(patterns) // 2, "--track")

        subprocess.run(args, cwd=root)

        written_patterns = set(Manifest.of(root).patterns)

        assert patterns.intersection(written_patterns) == patterns

    def test_passing_in_ignores(self, root):
        patterns = {"scoot", "flute"}
        args = ["gsb", "init"]
        for pattern in patterns:
            args.extend(("--ignore", pattern))

        subprocess.run(args, cwd=root)

        written_patterns = set((root / ".gitignore").read_text().splitlines())

        assert patterns.intersection(written_patterns) == patterns

    def test_mixing_includes_and_ignores(self, root):
        subprocess.run(
            [
                "gsb",
                "init",
                "toot",
                "--ignore",
                "flute",
                "--track",
                "boot",
                "*oot/**",
                "--ignore",
                "scoot",
                "--path",
                str(root.absolute()),
            ]
        )

        includes = set(Manifest.of(root).patterns)
        ignores = set((root / ".gitignore").read_text().splitlines())

        expected_includes = {"toot", "boot", "*oot/**"}
        expected_ignores = {"flute", "scoot"}
        assert expected_includes.intersection(includes) == expected_includes
        assert expected_ignores.intersection(ignores) == expected_ignores
        assert expected_includes.intersection(ignores) == set()
        assert expected_ignores.intersection(includes) == set()



"""Functionality for onboarding a new save state"""
from pathlib import Path
from typing import Iterable

from . import _git, backup
from .manifest import MANIFEST_NAME, Manifest

DEFAULT_IGNORE_LIST: tuple[str, ...] = ()


def create_repo(
    repo_root: Path,
    *patterns: str,
    ignore: Iterable[str] | None = None,
    name: str | None = None,
) -> Manifest:
    """Create a new `gsb`-managed git repo in the specified location

    Parameters
    ----------
    repo_root : Path
        The directory where the repo should be created
    patterns : str
        List of glob-patterns to match, specifying what in the working directory
        should be archived. If none are provided, then it will be assumed that
        the intent is to back up the *entire* folder and all its contents.
    ignore : list of str, optional
        List of glob-patterns to *ignore*. If None are specified, then nothing
        will be ignored.
    name : str, optional
        An alias to assign to the repo. If None is provided, the `repo_root`
        folder's name will be used.

    Returns
    -------
    Manifest
        The static configuration for that repo

    Raises
    ------
    FileExistsError
        If there is already a `gsb` repo in that location
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    if (repo_root / MANIFEST_NAME).exists():
        raise FileExistsError(f"{repo_root} already contains a GSB-managed repo")
    if not patterns:
        patterns = (".",)
    if "." not in patterns:
        patterns = tuple(set(patterns))

    _git.init(repo_root)

    _update_gitignore(repo_root, ignore or ())

    # enforce round-trip
    Manifest(repo_root, name or repo_root.resolve().name, tuple(patterns)).write()
    manifest = Manifest.of(repo_root)

    backup.create_backup(repo_root, "Start of gsb tracking")

    return manifest


def _update_gitignore(repo_root: Path, patterns: Iterable[str]) -> None:
    """Create or append to the ".gitignore" file in the specified repo

    Parameters
    ----------
    repo_root : Path
        The directory where the repo should be created
    patterns
        List of glob-patterns to ignore
    """
    with open(repo_root / ".gitignore", "a+", encoding="UTF-8") as gitignore:
        gitignore.seek(0)
        existing_lines: list[str] = [line.strip() for line in gitignore.readlines()]
        if existing_lines and "# gsb" not in existing_lines:
            gitignore.write("\n# gsb\n")
        new_lines: list[str] = sorted(
            [
                pattern
                for pattern in {*DEFAULT_IGNORE_LIST, *patterns}
                if pattern not in existing_lines
            ]
        )
        gitignore.write("\n".join(new_lines) + "\n")



"""Functionality for creating backups"""
import datetime as dt
import logging
from pathlib import Path

from . import _git
from .logging import IMPORTANT
from .manifest import MANIFEST_NAME, Manifest

REQUIRED_FILES: tuple[Path, ...] = (Path(".gitignore"), Path(MANIFEST_NAME))

LOGGER = logging.getLogger(__name__)


def generate_tag_name() -> str:
    """Generate a new calver-ish tag name

    Returns
    -------
    str
        A tag name that should hopefully be distinctly GSB and distinct
        from any tags a user would manually create

    Notes
    -----
    When unit testing, this method will need to be monkeypatched to provide
    even greater granularity... unless you want to use `time.sleep(1)` between
    each tag :O
    """
    return dt.datetime.now().strftime("gsb%Y.%m.%d+%H%M%S")


def create_backup(
    repo_root: Path,
    tag_message: str | None = None,
    commit_message: str | None = None,
    parent: str | None = None,
    tag_name: str | None = None,
) -> str:
    """Create a new backup

    Parameters
    ----------
    repo_root : Path
        The directory containing the GSB-managed repo
    tag_message : str, optional
        By default, this method just creates an "untagged" backup with a default
        commit message. To tag this backup, provide a description of the backup
        to use for both the commit message and the tag annotation.
    commit_message : str, optional
        By default, the commit message will match the `tag_message`, if one
        is provided. Provide a value to this argument to explicitly set the
        commit message. If neither a `tag` nor a `commit_message` is provided,
        the default value will be "GSB-managed commit"
    parent: str, optional
        By default, this new backup will be created as an incremental commit
        off of the current head. To instead reset to a specific revision,
        pass in an ID for that revision to this argument.
    tag_name : str, optional
        By default, tag names are automatically generated. Use this argument to
        provide a custom tag name. This option is ignored when not creating
        a tag.

    Returns
    -------
    str
        An identifier for the backup in the form of either a commit hash or a
        tag name

    Raises
    ------
    OSError
        If the specified repo does not exist or is not a gsb-managed repo
    ValueError
        If there are no changes to commit and no tag message was provided, or
        if the specified "parent" could not be resolved.
    """
    manifest = Manifest.of(repo_root)

    if parent is not None:
        _git.reset(repo_root, parent, hard=False)

    _git.add(repo_root, manifest.patterns)
    _git.force_add(repo_root, REQUIRED_FILES)
    try:
        identifier = _git.commit(
            repo_root, commit_message or tag_message or "GSB-managed commit"
        ).hash
        LOGGER.info("Changes committed with hash %s", identifier[:8])
        LOGGER.debug("Full hash: %s", identifier)
    except ValueError:
        if not tag_message:
            raise
        LOGGER.info("Nothing new to commit--all files are up-to-date.")
    if tag_message:
        head = _git.show(repo_root, "HEAD")
        for tag in _git.get_tags(repo_root, annotated_only=True):
            if tag.target == head:
                raise ValueError(f"The current HEAD is already tagged as {tag.name}")
        identifier = _git.tag(
            repo_root, tag_name or generate_tag_name(), tag_message
        ).name
        LOGGER.log(IMPORTANT, "Created new tagged backup: %s", identifier)
    return identifier



"""Common fixtures for use across the test package"""
import datetime as dt
import shutil
import time
from typing import Generator

import pytest

from gsb import _git, backup, history
from gsb.manifest import MANIFEST_NAME, Manifest


@pytest.fixture(autouse=True)
def suppress_git_config(monkeypatch):
    def empty_git_config() -> dict[str, str]:
        return {}

    monkeypatch.setattr(_git, "_git_config", empty_git_config)


@pytest.fixture
def patch_tag_naming(monkeypatch):
    def tag_name_generator() -> Generator[str, None, None]:
        date = dt.date(2023, 7, 10)
        while True:
            yield date.strftime("gsb%Y.%m.%d")
            date += dt.timedelta(days=1)

    tag_namer = tag_name_generator()

    def mock_tag_namer() -> str:
        return next(tag_namer)

    monkeypatch.setattr(backup, "generate_tag_name", mock_tag_namer)


@pytest.fixture(scope="session")
def _repo_with_history(tmp_path_factory):
    root = tmp_path_factory.mktemp("saves") / "fossil record"
    root.mkdir()
    _git._repo(root, new=True, initial_branch="main")

    (root / ".touched").touch()
    _git.add(root, [".touched"])

    _git.commit(root, "First commit", _committer=("you-ser", "me@computer"))
    _git.tag(root, "Init", None, _tagger=("you-ser", "me@computer"))

    (root / "species").write_text("trilobite\n")
    _git.add(root, ["species"])
    _git.commit(root, "Add an animal", _committer=("you-ser", "me@computer"))

    with (root / "species").open("a") as f:
        f.write("hallucigenia\n")

    _git.add(root, ["species"])
    _git.commit(root, "I think I'm drunk", _committer=("you-ser", "me@computer"))
    _git.tag(
        root,
        "0.1",
        "Cambrian period",
        _tagger=("you-ser", "me@computer"),
    )

    (root / "species").write_text("trilobite\n")
    _git.add(root, ["species"])
    _git.commit(root, "Remove hallucigenia", _committer=("you-ser", "me@computer"))
    _git.tag(
        root,
        "0.2",
        "Hello Permian period",
        _tagger=("you-ser", "me@computer"),
    )

    (root / "species").unlink()
    _git.add(root, ["species"])
    _git.commit(root, "Oh no! Everyone's dead!", _committer=("you-ser", "me@computer"))

    Manifest(root, "history of life", ("species",)).write()
    (root / ".gitignore").touch()
    _git.add(root, ["species", MANIFEST_NAME, ".gitignore"])
    _git.commit(root, "Start of GSB tracking")
    _git.tag(root, "gsb1.0", "Start of gsb tracking")

    (root / "species").write_text(
        "\n".join(("ichthyosaurs", "archosaurs", "plesiosaurs", "therapsids")) + "\n"
    )

    _git.add(root, ["species"])
    _git.commit(root, "Autocommit")
    _git.tag(root, "gsb1.1", "Triassic")

    time.sleep(1)

    (root / "species").write_text("plesiosaurs\n")
    _git.add(root, ["species"])
    jurassic = _git.commit(root, "Autocommit")

    (root / "species").write_text(
        "\n".join(("sauropods", "therapods", "plesiosaurs", "pterosaurs", "squids"))
        + "\n"
    )

    _git.add(root, ["species"])
    _git.commit(root, "Autocommit")
    _git.tag(root, "gsb1.2", "Jurassic")

    (root / "species").write_text(
        "\n".join(
            (
                "sauropods",
                "therapods",
                "raptors",
                "pliosaurs",
                "plesiosaurs",
                "mosasaurs",
                "pterosaurs",
            )
        )
        + "\n"
    )

    _git.add(root, ["species"])
    _git.commit(root, "Autocommit")

    with (root / "species").open("a") as f:
        f.write("mammals\n")

    _git.add(root, ["species"])
    _git.commit(root, "It's my ancestors!", _committer=("you-ser", "me@computer"))

    with (root / "species").open("a") as f:
        f.write("\n".join(("birds", "insects", "sharks", "squids")) + "\n")

    _git.add(root, ["species"])
    _git.commit(root, "Autocommit")

    _git.tag(root, "gsb1.3", "Cretaceous (my gracious!)")

    (root / "continents").write_text(
        "\n".join(
            (
                "laurasia",
                "gondwana",
            )
        )
        + "\n"
    )
    Manifest.of(root)._replace(patterns=("species", "continents", "oceans")).write()
    _git.add(root, ("continents",))
    _git.force_add(root, (MANIFEST_NAME,))
    _git.commit(root, "Autocommit")

    # these contents are new since the last commit
    (root / "oceans").write_text(
        "\n".join(
            (
                "pacific",
                "tethys",
            )
        )
        + "\n"
    )

    yield root, jurassic.timestamp


@pytest.fixture
def root(_repo_with_history, tmp_path):
    """Because the repo-with-history setup is so expensive, we want to perform our
    tests (some of which may rewrite the history) on a copy"""
    destination = tmp_path / "cloney"
    shutil.copytree(_repo_with_history[0], destination)
    yield destination


@pytest.fixture(scope="session")
def jurassic_timestamp(_repo_with_history):
    yield _repo_with_history[1]


@pytest.fixture(scope="session")
def all_backups(_repo_with_history):
    yield history.get_history(
        _repo_with_history[0], tagged_only=False, include_non_gsb=True
    )



"""Tests for rewriting repo histories"""
import logging
import subprocess

import pytest

from gsb import _git, fastforward
from gsb.backup import create_backup
from gsb.history import get_history
from gsb.rewind import restore_backup


class TestDeleteBackups:
    def test_deleting_a_backup(self, root):
        fastforward.delete_backups(root, "gsb1.1")
        assert [revision["identifier"] for revision in get_history(root)] == [
            "gsb1.3",
            "gsb1.2",
            "gsb1.0",
        ]

    def test_modified_files_are_automatically_backed_up(self, root):
        fastforward.delete_backups(root, "gsb1.3")

        # make sure the backup was deleted
        assert [revision["identifier"] for revision in get_history(root, limit=2)] == [
            "gsb1.2",
            "gsb1.1",
        ]

        # make sure there's nothing to commit post-ff
        _git.add(root, ["oceans"])
        with pytest.raises(ValueError, match="Nothing to commit"):
            _git.commit(root, "Oh no! Continents weren't being tracked!")

        # make sure that the unsaved contents were backed up (and preserved)
        assert [
            line.strip() for line in (root / "oceans").read_text("utf-8").splitlines()
        ] == [
            "pacific",
            "tethys",
        ]

    @pytest.mark.usefixtures("patch_tag_naming")
    def test_deleting_a_backup_doesnt_mess_up_subsequent_backups(self, root):
        fastforward.delete_backups(root, "gsb1.1")
        restore_backup(root, "gsb1.2")  # TODO: replace with export-backup
        assert (root / "species").read_text() == "\n".join(
            (
                "sauropods",
                "therapods",
                "plesiosaurs",
                "pterosaurs",
                "squids",
            )
        ) + "\n"

    @pytest.mark.usefixtures("patch_tag_naming")
    def test_deleting_a_backup_preserves_subsequent_backup_timestamps(
        self, root, jurassic_timestamp
    ):
        fastforward.delete_backups(root, "gsb1.0")
        assert [
            revision["identifier"]
            for revision in get_history(root, since=jurassic_timestamp)
        ] == ["gsb1.3", "gsb1.2"]

    def test_deleting_multiple_backups(self, root, all_backups):
        _git.reset(root, "gsb1.3", hard=True)

        # frequent workflow: deleting all non-tagged backups
        fastforward.delete_backups(
            root,
            *(
                revision["identifier"]
                for revision in all_backups[1:-1]
                if not revision["identifier"].startswith(("gsb", "0."))
            ),
        )
        assert [
            backup["identifier"] for backup in get_history(root, tagged_only=False)
        ] == [
            "gsb1.3",
            "gsb1.2",
            "gsb1.1",
            "gsb1.0",
            "0.2",
            "0.1",
        ]

    def test_raise_value_error_on_invalid_backup(self, root):
        with pytest.raises(ValueError, match=r"^Could not find((.|\n)*)gsb1.4"):
            fastforward.delete_backups(root, "gsb1.4")

    @pytest.mark.xfail(reason="not implemented")
    def test_deleting_the_very_first_backup(self, root, all_backups):
        fastforward.delete_backups(root, all_backups[-1]["identifier"])

        # can't check strict equality because commit hashes will have changed
        assert (
            len(get_history(root, tagged_only=False, include_non_gsb=True))
            == len(all_backups) - 1
        )

    def test_branch_post_ff_is_gsb(self, root, caplog):
        fastforward.delete_backups(root, "gsb1.0")
        assert _git._repo(root).head.shorthand == "gsb"
        assert "Could not delete branch" in "\n".join(
            [
                record.message
                for record in caplog.records
                if record.levelno == logging.WARNING
            ]
        )

    def test_original_non_gsb_branch_is_not_deleted(self, root):
        fastforward.delete_backups(root, "0.2")
        assert "main" in _git._repo(root).branches.local

    def test_cant_delete_backup_from_outside_history(self, root):
        _git.checkout_branch(root, "gsb", "gsb1.1")
        with pytest.raises(ValueError, match="not within the linear commit history"):
            fastforward.delete_backups(root, "gsb1.2")


class TestCLI:
    def test_no_args_initiates_prompt_in_cwd(self, root):
        result = subprocess.run(
            ["gsb", "delete"],
            cwd=root,
            capture_output=True,
            input="q\n".encode(),
        )

        assert (
            "Select a revision or revisions"
            in result.stdout.decode().strip().splitlines()[-1]
        )

    def test_prompt_includes_all_commits_since_last_tag(self, root, all_backups):
        post_tag_backup_1 = all_backups[0]["identifier"]
        post_tag_backup_2 = create_backup(root)[:8]

        result = subprocess.run(
            ["gsb", "delete"],
            cwd=root,
            capture_output=True,
            input="q\n".encode(),
        )
        assert [
            f"- {post_tag_backup_2}",
            f"- {post_tag_backup_1}",
            "- gsb1.3",
        ] == [
            line.split("from")[0].strip()
            for line in result.stderr.decode().strip().splitlines()[1:4]
        ]

    def test_passing_in_a_custom_root(self, root):
        _ = subprocess.run(
            ["gsb", "delete", "--path", root.name, "0.2"],
            cwd=root.parent,
            capture_output=True,
            input="q\n".encode(),
        )

        assert (
            get_history(root, tagged_only=True, include_non_gsb=True)[-2]["identifier"]
            == "gsb1.0"
        )

    def test_deleting_tag_by_argument(self, root):
        _ = subprocess.run(["gsb", "delete", "gsb1.1"], cwd=root, capture_output=True)

        assert [
            backup["identifier"]
            for backup in get_history(root, tagged_only=True, limit=3)
        ] == [
            "gsb1.3",
            "gsb1.2",
            "gsb1.0",
        ]

    @pytest.mark.parametrize(
        "how",
        (
            "short",
            "full",
        ),
    )
    def test_deleting_by_commit(self, root, how, all_backups):
        _git.reset(root, "gsb1.3", hard=True)  # reset to just after tagged bkp
        some_backup = all_backups[2]

        # meta-test to make sure I didn't grab a tag
        assert not some_backup["tagged"]

        if how == "short":
            some_commit = some_backup["identifier"]
        else:
            some_commit = some_backup["commit_hash"]

        _ = subprocess.run(
            ["gsb", "delete", some_commit],
            cwd=root,
            capture_output=True,
        )

        assert [
            backup["identifier"]
            for backup in get_history(root, tagged_only=False, limit=3)[::2]
        ] == [
            "gsb1.3",
            "gsb1.2",
        ]

    def test_deleting_by_prompt(self, root):
        _ = subprocess.run(
            ["gsb", "delete"],
            cwd=root,
            capture_output=True,
            input="gsb1.1\n".encode(),
        )

        assert [
            backup["identifier"] for backup in get_history(root, tagged_only=True)
        ] == [
            "gsb1.3",
            "gsb1.2",
            "gsb1.0",
        ]

    @pytest.mark.parametrize("how", ("by_argument", "by_prompt"))
    def test_unknown_revision_raises_error(self, root, how):
        arguments = ["gsb", "delete"]
        response = ""
        if how == "by_argument":
            arguments.append("gsb1.4")
        else:  # if how == "by_prompt"
            response = "gsb1.4\n"

        result = subprocess.run(
            arguments,
            cwd=root,
            capture_output=True,
            input=response.encode(),
        )

        assert result.returncode == 1
        assert "gsb1.4" in result.stderr.decode().strip().splitlines()[-1]

    @pytest.mark.parametrize("how", ("by_argument", "by_prompt"))
    def test_multi_delete(self, root, how):
        arguments = ["gsb", "delete"]
        response = ""
        if how == "by_argument":
            arguments.extend(["gsb1.0", "gsb1.1", "gsb1.2"])
        else:  # if how == "by_prompt"
            response = "gsb1.0, gsb1.1,gsb1.2\n"

        _ = subprocess.run(
            arguments,
            cwd=root,
            capture_output=True,
            input=response.encode(),
        )

        assert [
            backup["identifier"]
            for backup in get_history(root, tagged_only=True, include_non_gsb=True)
        ] == ["gsb1.3", "0.2", "0.1"]

    def test_running_on_repo_with_no_tags_retrieves_gsb_commits(self, tmp_path):
        """Like, I guess if the user deleted the initial backup"""
        repo = tmp_path / "repo"
        repo.mkdir()
        something = repo / "file"
        something.touch()
        _git.init(repo)
        _git.add(repo, [something.name])
        commit_hash = _git.commit(repo, "Hello").hash[:8]

        result = subprocess.run(
            ["gsb", "delete"], cwd=repo, capture_output=True, input="q\n".encode()
        )
        assert f"{commit_hash}" in result.stderr.decode().strip().splitlines()[1]

    def test_running_on_non_gsb_prompts_with_git_commits(self, tmp_path):
        repo = tmp_path / "repo"
        repo.mkdir()
        something = repo / "file"
        something.touch()
        _git.init(repo)
        _git.add(repo, [something.name])
        commit_hash = _git.commit(repo, "Hello", _committer=("Testy", "Testy")).hash[:8]

        result = subprocess.run(
            ["gsb", "delete"], cwd=repo, capture_output=True, input="q\n".encode()
        )
        log_lines = result.stderr.decode().strip().splitlines()

        assert "No GSB revisions found" in log_lines[1]
        assert f"{commit_hash}" in log_lines[2]

    def test_running_on_empty_repo_raises(self, tmp_path):
        repo = tmp_path / "repo"
        repo.mkdir()
        something = repo / "file"
        something.touch()
        _git.init(repo)

        result = subprocess.run(["gsb", "delete"], cwd=repo, capture_output=True)
        assert result.returncode == 1
        assert "No revisions found" in result.stderr.decode().strip().splitlines()[-1]

    def test_deleting_tells_you_to_run_git_gc_when_done(self, root):
        result = subprocess.run(
            ["gsb", "delete", "gsb1.1"], cwd=root, capture_output=True
        )

        assert "git gc" in result.stderr.decode().strip().splitlines()[-1]



"""Tests for reviewing repo histories"""
import subprocess

import pytest

from gsb import history
from gsb.backup import create_backup


class TestGetHistory:
    @pytest.mark.parametrize("create_root", (True, False), ids=("no_git", "no_folder"))
    def test_raises_when_theres_no_git_repo(self, tmp_path, create_root):
        random_folder = tmp_path / "random folder"
        if create_root:
            random_folder.mkdir()
        with pytest.raises(OSError):
            history.get_history(random_folder)

    def test_get_history_by_default_returns_all_gsb_tags(self, root):
        assert [revision["identifier"] for revision in history.get_history(root)] == [
            "gsb1.3",
            "gsb1.2",
            "gsb1.1",
            "gsb1.0",
        ]

    def test_get_history_can_limit_the_number_of_revisions(self, root):
        assert [
            revision["identifier"] for revision in history.get_history(root, limit=1)
        ] == ["gsb1.3"]

    def test_get_history_can_limit_revisions_by_date(self, root, jurassic_timestamp):
        assert [
            revision["identifier"]
            for revision in history.get_history(root, since=jurassic_timestamp)
        ] == ["gsb1.3", "gsb1.2"]

    def test_get_history_can_return_interim_commits_as_well(self, root):
        assert [
            revision["description"]
            for revision in history.get_history(root, tagged_only=False)
        ] == [
            "Autocommit",
            "Cretaceous (my gracious!)",
            "Autocommit",
            "Jurassic",
            "Autocommit",
            "Triassic",
            "Start of gsb tracking",
        ]

    def test_get_history_can_return_non_gsb_tags_as_well(self, root):
        assert [
            revision["identifier"]
            for revision in history.get_history(root, include_non_gsb=True)
        ] == [
            "gsb1.3",
            "gsb1.2",
            "gsb1.1",
            "gsb1.0",
            "0.2",
            "0.1",
        ]

    def test_get_history_can_return_non_gsb_commits_as_well(self, root):
        assert [
            revision["description"]
            for revision in history.get_history(
                root, tagged_only=False, include_non_gsb=True, limit=3
            )
        ] == [
            "Autocommit",
            "Cretaceous (my gracious!)",
            "It's my ancestors!",
        ]

    def test_getting_revisions_since_last_tagged_backup(self, root):
        create_backup(root)
        assert (
            len(
                history.get_history(
                    root, tagged_only=False, since_last_tagged_backup=True
                )
            )
            == 2
        )

    def test_always_include_latest_ignores_limit(self, root, all_backups):
        assert (
            history.get_history(root, limit=0, always_include_latest=True)
            == all_backups[0:1]
        )


class TestCLI:
    @pytest.fixture
    def last_backup(self, all_backups):
        assert not all_backups[0]["tagged"]  # meta-test
        yield f'1. {all_backups[0]["identifier"]}'

    def test_default_options_returns_all_gsb_tags_for_the_cwd(self, root):
        result = subprocess.run(["gsb", "history"], cwd=root, capture_output=True)
        backups = [
            line.split(" from ")[0] for line in result.stderr.decode().splitlines()
        ][1:]
        assert backups == ["2. gsb1.3", "3. gsb1.2", "4. gsb1.1", "5. gsb1.0"]

    def test_most_recent_backup_is_always_at_the_top(self, root, last_backup):
        result = subprocess.run(["gsb", "history"], cwd=root, capture_output=True)
        assert (
            result.stderr.decode().splitlines()[0].split(" from ")[0].strip()
            == last_backup
        )

    def test_even_a_non_gsb_backup_will_be_shown(self, root, last_backup):
        commit_hash = create_backup(root)
        result = subprocess.run(["gsb", "history"], cwd=root, capture_output=True)
        assert (
            result.stderr.decode().splitlines()[0].split(" from ")[0].strip()
            == f"1. {commit_hash[:8]}"
        )

    @pytest.mark.parametrize("how", ("by_argument", "by_option"))
    def test_passing_in_a_custom_root(self, root, how):
        args = ["gsb", "history", root]
        if how == "by_option":
            args.insert(2, "--path")

        result = subprocess.run(args, capture_output=True)

        assert result.stderr.decode().splitlines()[1].strip().startswith("2. gsb1.3")

    @pytest.mark.parametrize("flag", ("--limit", "-n", "-n2"))
    def test_setting_a_limit(self, root, last_backup, flag):
        args = ["gsb", "history", flag]
        if flag != "-n2":
            args.append("2")

        result = subprocess.run(args, capture_output=True, cwd=root)

        backups = [
            line.split(" from ")[0] for line in result.stderr.decode().splitlines()
        ]
        assert backups == [last_backup, "2. gsb1.3"]

    @pytest.mark.parametrize("flag", ("--limit", "-n", "-n0"))
    def test_raise_when_an_invalid_limit_is_set(self, root, flag):
        args = ["gsb", "history", flag]
        if flag != "-n0":
            args.append("-1")

        result = subprocess.run(args, capture_output=True, cwd=root)

        assert "Limit must be a positive integer" in result.stderr.decode()

    def test_setting_since(self, root, jurassic_timestamp, last_backup):
        args = ["gsb", "history", "--since", jurassic_timestamp.isoformat()]

        result = subprocess.run(args, capture_output=True, cwd=root)

        backups = [
            line.split(" from ")[0] for line in result.stderr.decode().splitlines()
        ]
        assert backups == [last_backup, "2. gsb1.3", "3. gsb1.2"]

    @pytest.mark.parametrize("flag", ("--include_non_gsb", "-g"))
    def test_including_non_gsb(self, root, last_backup, flag):
        args = ["gsb", "history", flag]

        result = subprocess.run(args, capture_output=True, cwd=root)

        backups = [
            line.split(" from ")[0] for line in result.stderr.decode().splitlines()
        ]
        assert backups == [
            last_backup,
            "2. gsb1.3",
            "3. gsb1.2",
            "4. gsb1.1",
            "5. gsb1.0",
            "6. 0.2",
            "7. 0.1",
        ]

    @pytest.mark.parametrize("flag", ("--all", "-a"))
    def test_including_commits(self, root, flag):
        args = ["gsb", "history", flag]

        result = subprocess.run(args, capture_output=True, cwd=root)

        backups = [
            line.split(" from ")[0]
            for line in [result.stderr.decode().splitlines()[i] for i in [1, 3, 5, 6]]
        ]
        assert backups == [
            "2. gsb1.3",
            "4. gsb1.2",
            "6. gsb1.1",
            "7. gsb1.0",
        ]



"""Tests for creating backups"""
import os
import subprocess
from pathlib import Path

import pygit2
import pytest

from gsb import _git, backup, onboard
from gsb.history import get_history
from gsb.manifest import Manifest


@pytest.fixture
def repo_root(tmp_path):
    root = tmp_path / "roto-rooter"

    my_world = root / "my world"
    my_world.mkdir(parents=True)

    my_save_data = my_world / "level.dat"
    my_save_data.write_text("Spawn Point: (0, 63, 0)\n")

    onboard.create_repo(root, my_world.name, ignore=["cruft", "ignore me"])

    my_save_data.write_text("Coordinates: (20, 71, -104)\n")

    (my_world / "new file").write_text("Hello, I'm new.\n")

    (my_world / "cruft").write_text("Boilerplate\n")

    ignoreme = my_world / "ignore me"
    ignoreme.mkdir()
    (ignoreme / "content.txt").write_text("Shouting into the void\n")

    yield root


@pytest.mark.usefixtures("patch_tag_naming")
class TestCreateBackup:
    @pytest.mark.parametrize("root_type", ("no_folder", "no_git", "no_manifest"))
    def test_raises_when_theres_no_gsb_repo(self, tmp_path, root_type):
        random_folder = tmp_path / "random folder"
        if root_type != "no_folder":
            random_folder.mkdir()
        if root_type == "no_manifest":
            _git.init(random_folder)
        with pytest.raises(OSError):
            backup.create_backup(random_folder)

    @pytest.mark.parametrize("tagged", (True, False), ids=("tagged", "untagged"))
    def test_backup_adds_from_manifest(self, repo_root, tagged):
        repo = _git._repo(repo_root, new=False)
        assert repo_root / "my world" / "new file" not in [
            Path(repo_root) / entry.path for entry in repo.index
        ]

        backup.create_backup(repo_root, tag_message="You're it" if tagged else None)

        repo = _git._repo(repo_root, new=False)
        assert repo_root / "my world" / "new file" in [
            Path(repo_root) / entry.path for entry in repo.index
        ]

    @pytest.mark.parametrize("tagged", (True, False), ids=("tagged", "untagged"))
    def test_backup_respects_gitignore(self, repo_root, tagged):
        backup.create_backup(repo_root, tag_message="You're it" if tagged else None)

        repo = _git._repo(repo_root, new=False)
        assert repo_root / "my world" / "ignore me" / "content.txt" not in [
            Path(repo_root) / entry.path for entry in repo.index
        ]

    def test_untagged_backup_is_a_commit(self, repo_root):
        identifier = backup.create_backup(repo_root)

        repo = _git._repo(repo_root, new=False)
        assert repo[identifier].type == pygit2.GIT_OBJ_COMMIT

    def test_tagged_backup_is_a_tag(self, repo_root):
        identifier = backup.create_backup(repo_root, "You're it")

        repo = _git._repo(repo_root, new=False)
        assert repo.revparse_single(identifier).type == pygit2.GIT_OBJ_TAG

    @pytest.mark.parametrize("tagged", (False, True), ids=("untagged", "tagged"))
    def test_raise_when_theres_nothing_new_to_backup(self, repo_root, tagged):
        backup.create_backup(repo_root, tag_message="You're it" if tagged else None)
        with pytest.raises(ValueError):
            backup.create_backup(
                repo_root, tag_message="You're still it" if tagged else None
            )

    def test_tagging_a_previously_untagged_backup(self, repo_root):
        commit_hash = backup.create_backup(repo_root)
        tag_name = backup.create_backup(repo_root, "You're it")

        repo = _git._repo(repo_root, new=False)
        assert str(repo.revparse_single(tag_name).target) == commit_hash

    def test_combining_with_previous_backup(self, root):
        identifier = backup.create_backup(root, parent="gsb1.3")
        assert [
            revision["identifier"]
            for revision in get_history(
                root, tagged_only=False, include_non_gsb=True, limit=2
            )
        ] == [identifier[:8], "gsb1.3"]

        # ensure that the squash kept all file changes
        assert (root / "continents").read_text("utf-8").startswith("laurasia")
        assert "oceans" in Manifest.of(root).patterns
        assert (root / "oceans").read_text("utf-8").strip().endswith("tethys")


class TestCLI:
    @pytest.fixture
    def prior_commits(self, repo_root):
        yield list(_git.log(repo_root))

    @pytest.fixture
    def prior_tags(self, repo_root):
        yield list(_git.get_tags(repo_root, False))

    def test_default_options_creates_untagged_backup_from_cwd(
        self, repo_root, prior_commits, prior_tags
    ):
        subprocess.run(["gsb", "backup"], cwd=repo_root)

        assert (
            len(list(_git.log(repo_root))),
            len(list(_git.get_tags(repo_root, False))),
        ) == (
            len(prior_commits) + 1,
            len(prior_tags),
        )

    @pytest.mark.parametrize("how", ("by_argument", "by_option"))
    def test_passing_in_a_custom_root(self, repo_root, how, prior_commits):
        args = ["gsb", "backup", repo_root]
        if how == "by_option":
            args.insert(2, "--path")

        subprocess.run(args)

        assert len(list(_git.log(repo_root))) == len(prior_commits) + 1

    @pytest.mark.usefixtures("patch_tag_naming")
    def test_creating_a_tagged_backup(self, repo_root, prior_commits, prior_tags):
        if os.name == "posix":
            subprocess.run(
                ['gsb backup --tag "Hello World"'], cwd=repo_root, shell=True
            )
        else:
            subprocess.run(["gsb", "backup", "--tag", "Hello World"], cwd=repo_root)

        tags = list(_git.get_tags(repo_root, False))

        assert (
            len(list(_git.log(repo_root))),
            len(tags),
        ) == (
            len(prior_commits) + 1,
            len(prior_tags) + 1,
        )

        assert tags[-1].annotation == "Hello World\n"

    def test_squash_with_last_commit(self, root):
        result = subprocess.run(["gsb", "backup", "-vc"], cwd=root, capture_output=True)
        commit_id = result.stderr.decode().splitlines()[-1].split("hash")[-1].strip()

        assert [
            revision["identifier"]
            for revision in get_history(
                root, tagged_only=False, include_non_gsb=True, limit=2
            )
        ] == [commit_id, "gsb1.3"]

    @pytest.mark.parametrize("confirm", (False, True), ids=("abort", "confirm"))
    def test_squash_with_last_tag_requires_confirmation(self, root, confirm):
        _git.reset(root, "gsb1.3", hard=False)

        result = subprocess.run(
            ["gsb", "backup", "-c"],
            cwd=root,
            capture_output=True,
            input=("y\n" if confirm else "\n").encode(),
        )

        if confirm:
            assert "gsb1.3" not in {
                tag.name for tag in _git.get_tags(root, annotated_only=False)
            }
        else:
            assert "Aborting" in result.stderr.decode().splitlines()[-1]

            assert (
                get_history(root, tagged_only=False, include_non_gsb=True, limit=1)[0][
                    "identifier"
                ]
                == "gsb1.3"
            )

    @pytest.mark.xfail(reason="not implemented")
    def test_squashing_with_the_first_backup(self, tmp_path):
        repo = tmp_path / "fresh"
        repo.mkdir()
        onboard.create_repo(repo, "something")
        result = subprocess.run(
            ["gsb", "backup", "-c"], cwd=repo, capture_output=True, input="y\n".encode()
        )
        assert result.returncode == 0
        assert len(get_history(repo, tagged_only=False)) == 1

    def test_squash_all_since_last_tag(self, root):
        # commit unsaved changes, so we can test that -cc works
        # even when all we're doing is squashing
        backup.create_backup(root)

        result = subprocess.run(
            ["gsb", "backup", "-vcc"], cwd=root, capture_output=True
        )
        commit_id = result.stderr.decode().splitlines()[-1].split("hash")[-1].strip()

        assert [
            revision["identifier"]
            for revision in get_history(
                root, tagged_only=False, include_non_gsb=True, limit=2
            )
        ] == [commit_id, "gsb1.3"]

    def test_squash_all_since_last_tag_is_quiet_when_theres_nothing_to_squash(
        self, root
    ):
        _git.reset(root, "gsb1.3", hard=False)

        result = subprocess.run(
            ["gsb", "backup", "-qcc"], cwd=root, capture_output=True
        )

        assert not result.stderr.decode()

        assert (
            get_history(root, tagged_only=False, include_non_gsb=True, limit=2)[-1][
                "identifier"
            ]
            == "gsb1.3"
        )

    def test_squash_all_since_last_tag_raises_when_there_are_no_tags(self, tmp_path):
        repo = tmp_path / "fresh"
        repo.mkdir()
        _git.init(repo)
        (repo / ".gitignore").touch()
        Manifest(repo, "blergh", ("something",)).write()
        backup.create_backup(repo)
        (repo / "something").write_text("it's not nothing\n")
        backup.create_backup(repo)
        result = subprocess.run(["gsb", "backup", "-cc"], cwd=repo, capture_output=True)
        assert result.returncode == 1
        assert len(get_history(repo, tagged_only=False)) == 2



"""Tests for exporting standalone backups"""
import subprocess
import tarfile
import zipfile
from pathlib import Path

import pytest

from gsb import _git, export
from gsb.manifest import Manifest


class TestExportBackup:
    @pytest.fixture(autouse=True)
    def put_autogenned_files_in_tmp(self, root, monkeypatch) -> None:
        original_archive_method = _git.archive

        def patched_archive(
            repo_root: Path, filename: Path, reference: str = "HEAD"
        ) -> None:
            if not filename.is_absolute():
                filename = root.parent / filename
            return original_archive_method(repo_root, filename, reference)

        monkeypatch.setattr(_git, "archive", patched_archive)

    def test_exporting_a_backup(self, root):
        export.export_backup(root, "gsb1.2")
        assert len(list(root.parent.glob("history of life_gsb1.2.*"))) == 1

    def test_exporting_a_backup_with_a_specific_filename(self, root, tmp_path):
        export.export_backup(root, "0.2", tmp_path / "exported.tbz")
        assert (tmp_path / "exported.tbz").exists()

    @pytest.mark.parametrize(
        "path_provided", (False, True), ids=("autogenned", "provided")
    )
    def test_exporting_will_not_overwrite_existing_file(
        self, root, tmp_path, path_provided
    ):
        (root.parent / "history of life_gsb1.0.zip").touch()
        (root.parent / "history of life_gsb1.0.tar.gz").touch()
        (tmp_path / "exported.tar.xz").touch()

        with pytest.raises(FileExistsError, match="already exists"):
            export.export_backup(
                root, "gsb1.0", tmp_path / "exported.tar.xz" if path_provided else None
            )

    def test_writing_zip_archive(self, root, tmp_path):
        export.export_backup(root, "gsb1.1", tmp_path / "exported.zip")
        with zipfile.ZipFile(tmp_path / "exported.zip") as archive:
            assert archive.read("species").decode("utf-8").startswith("ichthyosaurs\n")

    @pytest.mark.parametrize("compression", (None, "gz", "bz2", "xz"))
    def test_writing_tar_archive(self, root, tmp_path, compression, all_backups):
        archive_path = tmp_path / "exported.tar"
        if compression:
            archive_path = archive_path.with_suffix(f".tar.{compression}")

        export.export_backup(root, all_backups[0]["commit_hash"], archive_path)
        with tarfile.open(archive_path, f'r:{compression or ""}') as archive:
            assert archive.extractfile("continents").read().decode(
                "utf-8"
            ).strip().splitlines() == [
                "laurasia",
                "gondwana",
            ]

    def test_no_extension_raises_value_error(self, root, tmp_path):
        with pytest.raises(ValueError, match="does not specify an extension"):
            export.export_backup(root, "gsb1.2", tmp_path / "archive")

    def test_unknown_extension_raises_not_implemented_error(self, root, tmp_path):
        with pytest.raises(NotImplementedError):
            export.export_backup(
                root, "gsb1.2", tmp_path / "archive.tar.proprietaryext"
            )

    def test_repo_name_is_sanitized(self, tmp_path, root):
        repo = tmp_path / "diabolical"
        repo.mkdir()
        Manifest(repo, "I\\'m / soo ?evil?", ("doot",)).write()
        (repo / "doot").touch()
        _git.init(repo)
        _git.add(repo, (".gsb_manifest", "doot"))
        commit = _git.commit(repo, "Blahblah")
        export.export_backup(repo, commit.hash[:8])
        assert len(list(root.parent.glob(f"*evil*{commit.hash[:8]}.*"))) == 1


class TestCLI:
    def test_no_args_initiates_prompt_in_cwd(self, root):
        result = subprocess.run(
            ["gsb", "export"], cwd=root, capture_output=True, input="q\n".encode()
        )

        assert (
            "Select one by number or identifier"
            in result.stderr.decode().strip().splitlines()[-2]
        )

    def test_passing_in_a_custom_root(self, root):
        result = subprocess.run(
            ["gsb", "export", "--path", root.name],
            cwd=root.parent,
            capture_output=True,
            input="q\n".encode(),
        )

        assert (
            "Select one by number or identifier"
            in result.stderr.decode().strip().splitlines()[-2]
        )

    @pytest.mark.parametrize("how", ("arg", "prompt", "index"))
    def test_exporting_a_tag_by(self, root, how) -> None:
        args = ["gsb1.1"] if how == "arg" else []
        if how == "prompt":
            script: bytes | None = "gsb1.1\n".encode()
        elif how == "index":
            script = "3\n".encode()
        else:  # how == "arg"
            script = None

        _ = subprocess.run(
            ["gsb", "export", *args], cwd=root, capture_output=True, input=script
        )

        assert len(list(root.glob("history of life_gsb1.1.*"))) == 1

    @pytest.mark.parametrize("how", ("short", "full", "prompt"))
    def test_exporting_a_commit_by_(self, root, all_backups, how):
        assert not all_backups[0]["tagged"]
        some_commit = all_backups[0]["commit_hash"]

        if how != "full":
            some_commit = some_commit[:8]

        args = [some_commit] if how != "prompt" else []
        script = None if how != "prompt" else "0\n".encode()

        _ = subprocess.run(
            ["gsb", "export", *args],
            cwd=root,
            capture_output=True,
            input=script,
        )

        assert len(list(root.glob(f"history of life_{some_commit}.*"))) == 1

    @pytest.mark.parametrize("how", ("by_argument", "by_prompt"))
    def test_unknown_revision_raises_error(self, root, how):
        args = ["gsb", "export"]
        answers = [""]
        if how == "by_argument":
            args.append("not_a_thing")
        else:
            answers.insert(0, "not_a_thing")

        result = subprocess.run(
            args,
            cwd=root,
            capture_output=True,
            input="\n".join(answers).encode(),
        )

        assert result.returncode == 1
        assert "Could not find" in result.stderr.decode().strip().splitlines()[-1]

    def test_running_on_repo_with_no_tags_retrieves_gsb_commits(self, tmp_path):
        """Like, I guess if the user deleted the initial backup"""
        repo = tmp_path / "repo"
        repo.mkdir()
        something = repo / "file"
        something.touch()
        _git.init(repo)
        _git.add(repo, [something.name])
        commit_hash = _git.commit(repo, "Hello").hash[:8]

        result = subprocess.run(
            ["gsb", "export"], cwd=repo, capture_output=True, input="q\n".encode()
        )
        assert f"1. {commit_hash}" in result.stderr.decode().strip().splitlines()[1]

    def test_specifying_archive_name(self, root):
        subprocess.run(
            ["gsb", "export", "gsb1.0", "-o", "../archive.tar"],
            cwd=root,
            capture_output=True,
        )

        with tarfile.open(root.parent / "archive.tar", "r") as archive:
            assert (
                archive.extractfile(".gitignore").read().decode("utf-8").strip() == ""
            )

    def test_raises_if_archive_name_has_no_extension(self, root):
        result = subprocess.run(
            ["gsb", "export", "gsb1.0", "-o", "../archive"],
            cwd=root,
            capture_output=True,
        )

        assert result.returncode == 1
        assert "extension" in result.stderr.decode().splitlines()[-1]

    @pytest.mark.parametrize("filename", ("provided_filename", "autogen_filename"))
    @pytest.mark.parametrize("how", ("explicit", "short_flag"))
    @pytest.mark.parametrize(
        "archive_format", ("zip", "tar", "tar.gz", "tar.bz2", "tar.xz")
    )
    def test_specifying_an_archive_format(self, root, archive_format, how, filename):
        if how == "explicit":
            args = [f"--format=.{archive_format}"]
        else:
            args = [
                {
                    "zip": "-p",
                    "tar": "-t",
                    "tar.gz": "-z",
                    "tar.bz2": "-j",
                    "tar.xz": "-J",
                }[archive_format]
            ]

        if filename == "provided_filename":
            expected_path = root.parent / f"archive.{archive_format}"
            args.extend(["--output", "../archive"])
        else:
            expected_path = root / f"history of life_gsb1.3.{archive_format}"

        subprocess.run(
            ["gsb", "export", "gsb1.3", *args], cwd=root, capture_output=True
        )

        if archive_format == "zip":
            with zipfile.ZipFile(expected_path) as archive:
                assert (
                    archive.read("species")
                    .decode("utf-8")
                    .strip()
                    .splitlines()[-1]
                    .strip()
                    == "squids"
                )
        else:
            with tarfile.open(expected_path) as archive:
                assert (
                    archive.extractfile("species")
                    .read()
                    .decode("utf-8")
                    .strip()
                    .splitlines()[-1]
                    .strip()
                    == "squids"
                )

    @pytest.mark.parametrize(
        "flags",
        (
            ["-jJ"],
            ["-z", "-t"],
            ["-p", "--format", "zip"],
            pytest.param(
                ["-zz"],
                marks=pytest.mark.xfail(
                    reason="this would be nasty to implement in Click"
                ),
            ),
        ),
        ids=("short_combined", "short_separate", "short_and_long", "repeated_flag"),
    )
    def test_raises_if_multiple_formats_are_specified(self, root, flags):
        result = subprocess.run(
            ["gsb", "export", *flags, "0.2"],
            cwd=root,
            capture_output=True,
        )

        assert result.returncode == 1
        assert "conflicting" in result.stderr.decode().splitlines()[-1].lower()



"""ZIP-compatible `repo.write_archive`, stand-alone for easier upstream sharing.
As it is adapted directly from pygit2, this module is licensed under the
GNU Public License v2."""
import datetime as dt
import zipfile
from time import time

from pygit2 import GIT_FILEMODE_LINK, Commit, Index, Oid, Repository, Tree


def write_zip_archive(
    repo: Repository,
    treeish,
    archive: zipfile.ZipFile,
    timestamp: int | None = None,
    prefix: str = "",
) -> None:
    """Implementation of `repo.write_archive` that supports ZIP. Writes treeish
    into an archive.

    If no timestamp is provided and 'treeish' is a commit, its committer
    timestamp will be used. Otherwise the current time will be used.

    All path names in the archive are added to 'prefix', which defaults to
    an empty string.

    Parameters
    ----------
    repo: Repository
        The git repository
    treeish
        The treeish to write
    archive : zipfile.ZipFile
        An archive from the 'zipfile' module.
    timestamp : int, optional
        (Epoch) timestamp to use for the files in the archive.
    prefix : str, optional
        Extra prefix to add to the path names in the archive.

    Notes
    -----
    h/t to https://stackoverflow.com/a/18432983 for the example on converting
    between `TarInfo` and `ZipInfo`.

    Example
    -------
    >>> import tarfile, pygit2
    >>> from gsb import _git
    >>> repo = pygit2.Repository('.')
    >>> with zipfile.ZipFile('foo.zip', 'w') as archive:
    ...     _git.write_zip_archive(repo, repo.head.target, archive)
    """
    # Try to get a tree form whatever we got
    # Try to get a tree form whatever we got
    if isinstance(treeish, (str, Oid)):
        treeish = repo[treeish]

    tree = treeish.peel(Tree)

    # if we don't have a timestamp, try to get it from a commit
    if not timestamp:
        try:
            commit = treeish.peel(Commit)
            timestamp = commit.committer.time
        except Exception:
            pass

    # as a last resort, use the current timestamp
    if not timestamp:
        timestamp = int(time())

    datetime = dt.datetime.fromtimestamp(timestamp)
    zip_datetime = (
        # per https://docs.python.org/3/library/zipfile.html#zipfile.ZipInfo.date_time
        datetime.year,
        datetime.month,
        datetime.day,
        datetime.hour,
        datetime.minute,
        datetime.second,
    )

    index = Index()
    index.read_tree(tree)

    for entry in index:
        content = repo[entry.id].read_raw()
        info = zipfile.ZipInfo(prefix + entry.path)
        info.file_size = len(content)
        info.date_time = zip_datetime
        # info.uname = info.gname = "root"  # git's archive-zip.c does not
        if entry.mode == GIT_FILEMODE_LINK:
            raise NotImplementedError(
                "ZIP archives with symlinks are not currently supported."
                "\nSee: https://bugs.python.org/issue37921"
            )
        # per https://github.com/git/git/blob/3a06386e/archive-zip.c#L339
        info.external_attr = entry.mode << 16 if (entry.mode & 111) else 0
        archive.writestr(info, content)



"""Abstraction around the git library interface (to allow for easier backend swaps"""
import datetime as dt
import getpass
import logging
import re
import socket
import tarfile
import zipfile
from functools import partial
from pathlib import Path
from typing import Any, Generator, Iterable, NamedTuple, Self

import pygit2

from ._make_zip_archive import write_zip_archive

LOGGER = logging.getLogger(__name__)


def init(repo_root: Path) -> pygit2.Repository:
    """Initialize (or re-initialize) a git repo, equivalent to running
    `git init`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo

    Returns
    -------
    repo
        The initialized git repository

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    return _repo(repo_root, new=True)


def _repo(
    repo_root: Path, new: bool = False, initial_branch: str = "gsb"
) -> pygit2.Repository:
    """Load a git repository from the specified location

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    new : bool, optional
        By default, this method loads existing repositories. To initialize a new
        repo, pass in `new=True`
    initial_branch : str, optional
        By default, if a new repo is to be created, it will be given the initial
        branch name "gsb." To override this behavior (say, for testing) provide
        a different value to this argument.

    Returns
    -------
    repo
        The requested git repository

    Raises
    ------
    NotADirectoryError
        If `repo_root` is not a directory
    FileNotFoundError
        If `repo_root` does not exist or the repo is not a valid repository
        and `new=False`
    OSError
        If `repo_root` cannot otherwise be accessed
    """
    repo_root = repo_root.expanduser().resolve()
    if not repo_root.exists():
        raise FileNotFoundError(f"{repo_root} does not exist")
    if not repo_root.is_dir():
        raise NotADirectoryError(f"{repo_root} is not a directory")
    if new:
        LOGGER.debug(
            "git init %s --initial-branch=%s", repr(str(repo_root)), initial_branch
        )
        return pygit2.init_repository(repo_root, initial_head=initial_branch)
    try:
        return pygit2.Repository(repo_root)
    except pygit2.GitError as maybe_no_git:
        if "repository not found" in str(maybe_no_git).lower():
            raise FileNotFoundError(maybe_no_git) from maybe_no_git
        raise  # pragma: no cover


def _config() -> dict[str, str]:
    """Load the global git config and fill in any missing needed values

    Returns
    -------
    dict
        The user's global git config settings

    Notes
    -----
    Loading a repo-specific git config is not supported by this method
    """
    config = _git_config()
    LOGGER.debug("git config --global --list")
    config["user.name"] = config.get("user.name") or getpass.getuser()
    if "user.email" not in config:
        config["user.email"] = f"{getpass.getuser()}@{socket.gethostname()}"

    config["committer.name"] = "gsb"
    config["committer.email"] = "gsb@openbagtwo.github.io"
    return config


def _git_config() -> dict[str, str]:  # pragma: no cover
    """Separate encapsulation for the purposes of monkeypatching"""
    try:
        return {
            entry.name: entry.value for entry in pygit2.Config().get_global_config()
        }
    except OSError:
        return {}


def add(repo_root: Path, patterns: Iterable[str]) -> pygit2.Index:
    """Add files matching the given pattern to the repo, equivalent to running
    `git add <pattern>`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    patterns : list of str
        The glob patterns to match

    Returns
    -------
    index
        The updated git index

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    repo = _repo(repo_root)
    patterns = list(patterns)
    LOGGER.debug(
        "git add %s",
        " ".join([repr(pattern) for pattern in patterns]),
    )
    repo.index.add_all(patterns)
    repo.index.write()
    return repo.index


def force_add(repo_root: Path, files: Iterable[Path]) -> pygit2.Index:
    """Forcibly add specific files, overriding .gitignore, equivalent to running
    `git add <file> --force`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    files : list of paths
        The file paths to add, relative to the repo root

    Returns
    -------
    index
        The updated git index

    Raises
    ------
    FileNotFoundError
        If one of the specified paths does not exist
    IsADirectoryError
        If one of the specified paths is a directory
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    repo = _repo(repo_root)
    for path in files:
        try:
            LOGGER.debug("git add --force %s", repr(str(path)))
            repo.index.add(path)
        except OSError as maybe_file_not_found:  # pragma: no cover
            if "No such file or directory" in str(maybe_file_not_found):
                raise FileNotFoundError(maybe_file_not_found) from maybe_file_not_found
            raise  # pragma: no cover
        except pygit2.GitError as maybe_directory:  # pragma: no cover
            if "is a directory" in str(maybe_directory):
                raise IsADirectoryError(maybe_directory) from maybe_directory
    repo.index.write()
    return repo.index


class Commit(NamedTuple):
    """Commit metadata

    Attributes
    ----------
    hash : str
        The full commit hash
    message : str
        The commit message
    timestamp : dt.datetime
        The timestamp of the commit
    gsb : bool
        True if and only if the tag was created by `gsb`
    """

    hash: str
    message: str
    timestamp: dt.datetime
    gsb: bool

    @classmethod
    def from_pygit2(cls, commit_object: pygit2.Object) -> Self:
        """Resolve from a pygit2 object"""
        try:
            gsb = commit_object.committer.name == "gsb"
        except AttributeError:  # pragma: no cover
            gsb = False
        return cls(
            str(commit_object.id),
            commit_object.message,
            dt.datetime.fromtimestamp(commit_object.commit_time),
            gsb,
        )


def commit(
    repo_root: Path,
    message: str,
    timestamp: dt.datetime | None = None,
    _committer: tuple[str, str] | None = None,
) -> Commit:
    """Commit staged changes, equivalent to running `git commit -m <message>`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    message : str
        The commit message
    timestamp : dt.datetime, optional
        By default, commits are created using the current timestamp. Use this
        argument to provide a custom timestamp (as you would when calling
        `git commit --date <timestamp>`)
    _committer : (str, str) tuple, optional
        By default this method uses "gsb" as the committer. This should not
        be overridden outside of testing, but to do so, pass in both the
        username and email address.

    Returns
    -------
    commit
        The generated commit object

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    ValueError
        If the commit is empty ("nothing to do")
    """
    repo = _repo(repo_root)
    try:
        ref = repo.head.name
        parents = [repo.head.target]
    except pygit2.GitError as headless:
        if re.search(r"reference 'refs/heads/(.*)' not found", str(headless)):
            ref = "HEAD"
            parents = []
        else:
            raise  # pragma: no cover
    if not repo.status(untracked_files="no"):
        raise ValueError("Nothing to commit")

    if not message.endswith("\n"):
        message += "\n"

    config = _config()
    author = pygit2.Signature(config["user.name"], config["user.email"])
    signature_kwargs: dict[str, Any] = {}
    if timestamp is not None:
        signature_kwargs = {"time": int(timestamp.timestamp())}
    if _committer is None:
        committer = pygit2.Signature(
            config["committer.name"], config["committer.email"], **signature_kwargs
        )
    else:
        committer = pygit2.Signature(*_committer, **signature_kwargs)

    LOGGER.debug("git commit -m %s", repr(message))
    commit_id = repo.create_commit(
        ref, author, committer, message, repo.index.write_tree(), parents
    )
    return Commit.from_pygit2(repo[commit_id])


def log(
    repo_root: Path, starting_point: str | None = None
) -> Generator[Commit, None, None]:
    """Return metadata about commits such as you'd get by running `git log`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    starting_point : str, optional
        The default behavior is to walk backwards from the current repo HEAD.
        To specify a different starting point, provide an identifier for this
        variable.

    Returns
    -------
    iterable of commit
        The requested commits, returned lazily, in reverse-chronological order

    Raises
    ------
    ValueError
        If `starting_point` is provided and can't be resolved
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    repo = _repo(repo_root)
    try:
        if starting_point is not None:
            LOGGER.debug("git log %s", starting_point)
            head = _resolve_reference(starting_point, repo).id
        else:
            LOGGER.debug("git log")
            head = repo[repo.head.target].id
        for commit_object in repo.walk(head, pygit2.GIT_SORT_NONE):
            yield Commit.from_pygit2(commit_object)
    except pygit2.GitError as maybe_empty_history:
        if re.search("reference (.*) not found", str(maybe_empty_history)):
            # this is what pygit2 throws when there's no commits
            return
        raise  # pragma: no cover


def ls_files(repo_root: Path) -> list[Path]:
    """List the files in the index, similar to the output you'd get from
    running `git ls-files`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo

    Returns
    -------
    list of Path
        The files being tracked in this repo

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    repo = _repo(repo_root)
    LOGGER.debug("git ls-files")
    return [repo_root / file.path for file in repo.index]


class Tag(NamedTuple):
    """Tag metadata

    Attributes
    ----------
    name : str
        The name of the tag
    annotation : str or None
        The tag's annotation. If None, then this is a lightweight tag
    target : Commit
        The commit the tag is targeting
    gsb : bool or None
        True if the tagger was  `gsb`, False if it was created by
        someone / something else and None if it's a lightweight tag (which
        doesn't have a tagger)
    """

    name: str
    annotation: str | None
    target: Commit
    gsb: bool | None

    @classmethod
    def from_repo_reference(
        cls, reference: pygit2.Reference | str, repo: pygit2.Repository
    ) -> Self:
        """Parse the reference and resolve from the pygit2 object"""
        if isinstance(reference, str):
            tag_object = repo.revparse_single(reference)
        else:
            tag_object = repo.revparse_single(reference.name)
            reference = reference.shorthand

        if tag_object.type == pygit2.GIT_OBJ_TAG:
            try:
                gsb = tag_object.tagger.name == "gsb"
            except AttributeError:  # pragma: no cover
                gsb = False
            return cls(
                tag_object.name,
                tag_object.message,
                Commit.from_pygit2(repo[tag_object.target]),
                gsb,
            )
        if tag_object.type == pygit2.GIT_OBJ_COMMIT:
            return cls(reference, None, Commit.from_pygit2(tag_object), False)
        raise TypeError(  # pragma: no cover
            f"Don't know how to parse reference of type: {tag_object.type}"
        )


def tag(
    repo_root: Path,
    tag_name: str,
    annotation: str | None,
    target: str | None = None,
    _tagger: tuple[str, str] | None = None,
) -> Tag:
    """Create a tag at the current HEAD, equivalent to running
    `git tag [-am <annotation>]`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    tag_name : str
        The name to give the tag
    annotation : str or None
        The annotation to give the tag. If None is provided, a lightweight tag
        will be created
    target : str, optional
        The commit to assign the tag. If None is given, the current HEAD will
        be used.
    _tagger : (str, str) tuple, optional
        By default this method uses "gsb" as the tagger. This should not
        be overridden outside of testing, but to do so, pass in both the
        username and email address.

    Returns
    -------
    tag
        The generated tag object

    Raises
    ------
    ValueError
        If there is already a tag with the provided name, or if `target` is
        provided and can't be resolved
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    repo = _repo(repo_root)

    config = _config()
    if _tagger is None:
        tagger = pygit2.Signature(
            config["committer.name"],
            config["committer.email"],
        )
    else:
        tagger = pygit2.Signature(*_tagger)

    reference = _resolve_reference(target, repo).id if target else repo.head.target
    ref_short = str(reference)[:8] if target else "HEAD"

    if annotation:
        if not annotation.endswith("\n"):
            annotation += "\n"

        LOGGER.debug("git tag %s -am %s %s", tag_name, repr(annotation), ref_short)
        repo.create_tag(
            tag_name,
            reference,
            pygit2.GIT_OBJ_COMMIT,
            tagger,
            annotation,
        )
    else:
        LOGGER.debug("git tag %s %s", tag_name, ref_short)
        repo.create_reference(f"refs/tags/{tag_name}", reference)

    return Tag.from_repo_reference(tag_name, repo)

    # PSA: pygit2.AlreadyExistsError subclasses ValueError


def delete_tag(repo_root: Path, tag_name: str) -> None:
    """Delete a tag, equivalent to running `git tag -d <tag_name>`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    tag_name : str
        The name to the tag to delete

    Raises
    ------
    ValueError
        If there is no tag with the provided name
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    repo = _repo(repo_root)
    try:
        LOGGER.debug("git tag -d %s", tag_name)
        repo.references[f"refs/tags/{tag_name}"].delete()
    except KeyError as tag_not_found:
        raise ValueError(f"No such tag: {tag_name}") from tag_not_found


def get_tags(repo_root: Path, annotated_only: bool) -> list[Tag]:
    """List the repo's tags, similar to the output you'd get from
    running `git tag`, with the additional option of filtering out
    lightweight tags

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    annotated_only : bool
        Lightweight tags will be included if and only if this is `False`

    Returns
    -------
    list of Tag
        The requested list of tags, sorted in lexical order

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    """
    repo = _repo(repo_root)
    tags: list[Tag] = []
    LOGGER.debug("git tag")
    for reference in repo.references.iterator(pygit2.GIT_REFERENCES_TAGS):
        parsed_tag = Tag.from_repo_reference(reference, repo)
        if parsed_tag.annotation or not annotated_only:
            tags.append(parsed_tag)
    return sorted(tags)


def _resolve_reference(reference: str, repo: pygit2.Repository) -> pygit2.Object:
    """Attempt to resolve a reference

    Parameters
    ----------
    reference : str
        The reference to resolve
    repo : Repository
        The git repository

    Returns
    -------
    pygit2.Object
        The resolved reference

    Raises
    ------
    ValueError
        If the specified revision does not exist
    """
    try:
        LOGGER.debug("git show %s", reference)
        return repo.revparse_single(reference)
    except KeyError as no_rev:
        raise ValueError(
            f"Could not find a revision named {repr(reference)}"
        ) from no_rev


def show(repo_root: Path, reference: str) -> Commit | Tag:
    """Get information about a specified revision, similar to the output you'd
    get from running `git show <commit-hash-or-tag-name>`.

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    reference : str
        A unique descriptor of the tag or commit

    Returns
    -------
    Commit or Tag
        The requested tag or commit

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    ValueError
        If the specified revision does not exist
    """
    repo = _repo(repo_root)
    revision = _resolve_reference(reference, repo)
    if revision.type == pygit2.GIT_OBJ_TAG:
        return Tag.from_repo_reference(str(revision.id), repo)
    if revision.type == pygit2.GIT_OBJ_COMMIT:
        return Commit.from_pygit2(revision)
    raise TypeError(  # pragma: no cover
        f"Object of type {revision.type} is not a valid revision"
    )


def reset(repo_root: Path, reference: str, hard: bool) -> None:
    """Reset the repo to the specified revision, equivalent to running
    `git reset [--hard/--soft] <revision>`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    reference : str
        A unique descriptor of the tag or commit
    hard : bool
        If True, perform a hard reset. If False, perform a soft reset.

    Returns
    -------
    None

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    ValueError
        If the specified revision does not exist
    """
    repo = _repo(repo_root)

    # make sure revision exists
    reference = _resolve_reference(reference, repo).id

    LOGGER.debug(f"git reset --{'hard' if hard else 'soft'} %s", reference)
    repo.reset(reference, pygit2.GIT_RESET_HARD if hard else pygit2.GIT_RESET_SOFT)


def checkout_files(repo_root: Path, reference: str, paths: Iterable[Path]) -> None:
    """Check out the versions of the specified files that existed at the specified
    revision, equivalent to running
    `git reset <revision> -- <paths...> && git checkout <revision> -- <paths...>`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    reference : str
        A unique descriptor of the tag or commit
    paths : list of Paths
        The files to reset

    Returns
    -------
    None

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    ValueError
        If the specified revision does not exist
    """
    repo = _repo(repo_root)

    revision = _resolve_reference(reference, repo)
    if isinstance(revision, pygit2.Tag):
        return checkout_files(repo_root, str(revision.target), paths)

    paths = list(paths)

    for path in paths:
        LOGGER.debug("git reset %s -- %s", reference, repr(str(path)))
        try:
            repo.index.remove(path)
        except OSError:
            pass  # possible that the file no longer exists
        try:
            past_file = revision.tree[path]
            repo.index.add(pygit2.IndexEntry(path, past_file.id, past_file.filemode))
        except KeyError:
            pass  # possible that the file doesn't exist at the time of the revision

    repo.index.write()
    LOGGER.debug(
        "git checkout %s -- %s",
        reference,
        " ".join((repr(str(path)) for path in paths)),
    )
    repo.checkout(strategy=pygit2.GIT_CHECKOUT_FORCE, paths=paths)
    return None


def checkout_branch(repo_root: Path, branch_name: str, target: str | None) -> None:
    """Check out a branch, either new or existing, equivalent to calling
    `git checkout [-b] <branch_name> [<target>]`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    branch_name : str
        The name for the branch
    target : str or None
        When a reference is provided, this method will attempt to create a new
        branch at that reference point. When None is provided, this method will
        attempt to check out an existing branch at that branch's head.

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    ValueError
        If the specified `target` does not exist, if the `branch_name` is taken
        (when `target` is specified) or if the `branch_name` _does not_ exist
        (when `target=None`)
    """
    repo = _repo(repo_root)
    if target is not None:
        LOGGER.debug("git checkout -b %s %s", branch_name, target)
        reference = _resolve_reference(target, repo)
        if isinstance(reference, pygit2.Tag):
            reference = _resolve_reference(str(reference.target), repo)
        repo.branches.local.create(branch_name, reference)
    try:
        LOGGER.debug("git checkout %s", branch_name)
        repo.checkout(repo.branches.local[branch_name])
    except KeyError as no_such_branch:
        raise ValueError(no_such_branch) from no_such_branch


def delete_branch(repo_root: Path, branch_name: str) -> None:
    """Delete a branch, equivalent to running `git branch -D <branch_name>`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    branch_name : str
        The name of the branch

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed
    ValueError
        If the specified branch does not exist or if the specified branch is
        currently checked out
    """
    repo = _repo(repo_root)
    try:
        LOGGER.debug("git branch -D %s", branch_name)
        repo.branches.local.delete(branch_name)
    except KeyError as no_such_branch:
        raise ValueError(no_such_branch) from no_such_branch


def archive(repo_root: Path, filename: Path, reference: str = "HEAD") -> None:
    """Create a standalone archive containing the files in the repo at the
    current HEAD, equivalent to running `git archive -o <filename>`

    Parameters
    ----------
    repo_root : Path
        The root directory of the git repo
    filename : Path
        The full path to the archive's location, including its extension
    reference : str, optional
        A unique descriptor of the tag or commit to archive. If None is given,
        the default is to use the current HEAD.

    Raises
    ------
    OSError
        If `repo_root` does not exist, is not a directory or cannot be accessed,
        or if the specified `filename` already exists or cannot be written to.
    ValueError
        If the specified `target` does not exist or if the given filename
        does not have a valid extension
    NotImplementedError
        If the compression schema implied by the filename's extension is not
        supported
    """
    repo = _repo(repo_root)
    revision = _resolve_reference(reference, repo)
    if filename.exists():
        raise FileExistsError(f"Archive {filename} already exists.")

    LOGGER.debug("git archive -o %s %s", filename, reference)

    match tuple(suffix.lower() for suffix in filename.suffixes):
        case ():
            raise ValueError(f"Filename {filename} does not specify an extension.")
        case *_, ".tar":
            opener = partial(tarfile.open, mode="x:")
        case (*_, ".tgz") | (*_, ".tar", ".gz"):
            opener = partial(tarfile.open, mode="x:gz")
        case (*_, ".tbz2" | ".tbz") | (*_, ".tar", (".bz2" | ".bz")):
            opener = partial(tarfile.open, mode="x:bz2")
        case (*_, ".txz" | ".tlzma" | ".tlz") | (*_, ".tar", (".xz" | ".lzma" | ".lz")):
            opener = partial(tarfile.open, mode="x:xz")
        case (*_, ".zip"):
            with zipfile.ZipFile(
                filename, "w", compression=zipfile.ZIP_DEFLATED
            ) as zip_file:
                write_zip_archive(repo, revision, zip_file)
            return
        case _:
            raise NotImplementedError(f"{filename}: Archive format is not supported.")

    with opener(filename) as archive_file:
        repo.write_archive(revision, archive_file)



"""Common logging utilities"""
import logging

IMPORTANT = 25  # INFO logs that should be displayed even without "-v"
logging.addLevelName(IMPORTANT, "INFO")


class CLIFormatter(logging.Formatter):
    """Colorful formatter for the CLI

    h/t https://stackoverflow.com/a/56944256
    """

    grey = "\x1b[2;20m"
    yellow = "\x1b[33;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    FORMATS = {
        logging.DEBUG: grey + "%(message)s" + reset,
        logging.INFO: "%(message)s",
        IMPORTANT: "%(message)s",
        logging.WARNING: yellow + "%(message)s" + reset,
        logging.ERROR: bold_red + "%(message)s" + reset,
        logging.CRITICAL: bold_red + "%(message)s" + reset,
    }

    def format(self, record: logging.LogRecord) -> str:
        return logging.Formatter(self.FORMATS.get(record.levelno)).format(record)


def verbosity_to_log_level(verbosity: int) -> int:
    """Convert a verbosity level (number of `-v`s minus number of `-q`s) to
    a logging level

    Parameters
    ----------
    verbosity: int
        A verbosity level usually specified by the number of `-v` flags a user
        provides minus the number of `-q` flags. As a baseline, a verbosity of
        0 will set the level to handle all messages above (excluding) the
        INFO level.

    Returns
    -------
    int
        The corresponding log level that should be set

    Notes
    -----
    Technically the default logging level is set just high enough to exclude
    DEBUG by default. This allows us to capture intermediate log levels (read:
    `IMPORTANT`) at the `verbosity = -1` (`-q`) level.
    """
    return logging.INFO + 1 - 10 * verbosity



"""Command-line interface"""
import datetime as dt
import functools
import logging
import sys
from pathlib import Path
from typing import Any, Callable

import click

from . import _git, _version
from . import backup as backup_
from . import export as export_
from . import fastforward
from . import history as history_
from . import onboard
from . import rewind as rewind_
from .logging import IMPORTANT, CLIFormatter, verbosity_to_log_level
from .manifest import Manifest

LOGGER = logging.getLogger(__package__)


@click.group()
@click.help_option("--help", "-h")
@click.version_option(_version.get_versions()["version"], "--version", "-v", "-V")
def gsb():
    """CLI for managing incremental backups of your save states using Git!"""


def _subcommand_init(command: Callable) -> Callable:
    """Register a subcommand and add some standard CLI handling"""

    @functools.wraps(command)
    def wrapped(path: Path | None, verbose: int, quiet: int, *args, **kwargs) -> None:
        cli_handler = logging.StreamHandler()
        cli_handler.setFormatter(CLIFormatter())
        LOGGER.addHandler(cli_handler)

        log_level = verbosity_to_log_level(verbose - quiet)

        cli_handler.setLevel(log_level)

        # TODO: when we add log files, set this to minimum log level across all handlers
        LOGGER.setLevel(log_level)
        try:
            command((path or Path()).absolute(), *args, **kwargs)
        except (OSError, ValueError) as oh_no:
            LOGGER.error(oh_no)
            sys.exit(1)

    wrapped = click.option(
        "--path",
        type=Path,
        metavar="SAVE_PATH",
        help=(
            "Optionally specify the root directory containing your save data."
            " If no path is given, the current working directory will be used."
        ),
    )(wrapped)

    wrapped = click.option(
        "--verbose",
        "-v",
        count=True,
        help="Increase the amount of information that's printed.",
    )(wrapped)

    wrapped = click.option(
        "--quiet",
        "-q",
        count=True,
        help="Decrease the amount of information that's printed.",
    )(wrapped)
    return gsb.command()(wrapped)


@click.option(
    "--tag",
    type=str,
    help='Specify a description for this backup and "tag" it for future reference.',
    metavar='"MESSAGE"',
)
@click.option(
    "--combine",
    "-c",
    count=True,
    help=(
        "Combine this backup and the last backup,"
        " or use -cc to combine ALL backups since the last tagged backup."
    ),
)
@click.argument(
    "path_as_arg",
    type=Path,
    required=False,
    metavar="[SAVE_PATH]",
)
@_subcommand_init
def backup(repo_root: Path, path_as_arg: Path | None, tag: str | None, combine: int):
    """Create a new backup."""
    parent_hash = None
    if combine == 1:
        try:
            combine_me, parent = history_.get_history(
                repo_root, tagged_only=False, include_non_gsb=True, limit=2
            )
        except ValueError as probably_not_enough_values:
            if "not enough values to unpack" in str(probably_not_enough_values):
                LOGGER.error("Cannot combine with the very first backup.")
                sys.exit(1)
            raise probably_not_enough_values  # pragma: no-cover

        LOGGER.log(IMPORTANT, "Combining with %s", combine_me["identifier"])
        if combine_me["tagged"]:
            LOGGER.warning("Are you sure you want to overwrite a tagged backup?")
            history_.log_revision(combine_me, None)
            confirmed: bool = click.confirm(
                "",
                default=False,
                show_default=True,
            )
            if not confirmed:
                LOGGER.error("Aborting.")
                sys.exit(1)
            _git.delete_tag(repo_root, combine_me["identifier"])
        parent_hash = parent["identifier"]
    if combine > 1:
        try:
            last_tag = history_.get_history(repo_root, tagged_only=True, limit=1)[0]
        except IndexError:
            LOGGER.error("There are no previous tagged backups.")
            sys.exit(1)
        LOGGER.log(IMPORTANT, "Combining with the following backups:")
        combining = history_.show_history(
            repo_root,
            tagged_only=False,
            include_non_gsb=True,
            since_last_tagged_backup=True,
        )
        if not combining:
            LOGGER.log(IMPORTANT, "(no backups to combine)")
        parent_hash = last_tag["identifier"]

    backup_.create_backup(path_as_arg or repo_root, tag, parent=parent_hash)


@click.option(
    "--ignore",
    type=str,
    required=False,
    multiple=True,
    help=(
        "Provide a glob pattern to ignore. Each ignore pattern"
        ' must be prefaced with the "--ignore" flag.'
    ),
)
@click.option(
    "--track",
    type=str,
    required=False,
    multiple=True,
    help=(
        "Provide a glob pattern to track (note: arguments without any flag will"
        " also be treated as track patterns)."
    ),
)
@click.argument(
    "track_args", type=str, required=False, nargs=-1, metavar="[TRACK_PATTERN]..."
)
@_subcommand_init
def init(
    repo_root: Path,
    track_args: tuple[str, ...],
    track: tuple[str, ...],
    ignore: tuple[str, ...],
):
    """Start tracking a save."""
    onboard.create_repo(repo_root, *track_args, *track, ignore=ignore)


@click.option(
    "--include_non_gsb",
    "-g",
    is_flag=True,
    help="Include backups created directly with Git / outside of gsb.",
)
@click.option("--all", "-a", "all_", is_flag=True, help="Include non-tagged backups.")
@click.option(
    "--since",
    type=click.DateTime(),
    required=False,
    help="Only show backups created after the specified date.",
)
@click.option(
    "--limit",
    "-n",
    type=int,
    required=False,
    help="The maximum number of backups to return.",
)
@click.argument(
    "path_as_arg",
    type=Path,
    required=False,
    metavar="[SAVE_PATH]",
)
@_subcommand_init
def history(
    repo_root: Path,
    path_as_arg: Path | None,
    limit: int | None,
    since: dt.datetime | None,
    all_: bool,
    include_non_gsb: bool,
):
    """List the available backups, starting with the most recent."""

    kwargs: dict[str, Any] = {
        "tagged_only": not all_,
        "include_non_gsb": include_non_gsb,
    }
    if limit is not None:
        if limit <= 0:
            LOGGER.error("Limit must be a positive integer")
            sys.exit(1)
        kwargs["limit"] = limit
    if since is not None:
        kwargs["since"] = since

    history_.show_history(
        path_as_arg or repo_root, **kwargs, always_include_latest=True
    )


@click.option(
    "--include_gsb_settings",
    is_flag=True,
    help="Also revert the GSB configuration files (including .gitignore)",
)
@click.argument(
    "revision",
    type=str,
    required=False,
)
@_subcommand_init
def rewind(repo_root: Path, revision: str | None, include_gsb_settings: bool):
    """Restore a backup to the specified REVISION."""
    if revision is None:
        revision = _prompt_for_a_recent_revision(repo_root)
    try:
        rewind_.restore_backup(repo_root, revision, not include_gsb_settings)
    except ValueError as whats_that:
        LOGGER.error(whats_that)
        sys.exit(1)


def _prompt_for_a_recent_revision(repo_root) -> str:
    """Select a recent revision from a prompt"""
    LOGGER.log(IMPORTANT, "Here is a list of recent backups:")
    revisions = history_.show_history(repo_root, limit=10)
    if len(revisions) == 0:
        LOGGER.info("No tagged revisions found. Trying untagged.")
        revisions = history_.show_history(
            repo_root,
            limit=10,
            tagged_only=False,
        )
    if len(revisions) == 0:
        LOGGER.warning("No GSB revisions found. Trying Git.")
        revisions = history_.show_history(
            repo_root,
            limit=10,
            tagged_only=False,
            include_non_gsb=True,
        )
    if len(revisions) == 0:
        LOGGER.error("No revisions found!")
        sys.exit(1)
    LOGGER.log(IMPORTANT, "\nMost recent backup:")
    most_recent_backup = history_.show_history(
        repo_root, limit=1, always_include_latest=True, numbering=0
    )[0]

    LOGGER.log(
        IMPORTANT,
        "\nSelect one by number or identifier (or [q]uit and"
        " call gsb history yourself to get more revisions).",
    )

    choice: str = click.prompt(
        "Select a revision", default="q", show_default=True, type=str
    ).lower()

    if choice.lower().strip() == "q":
        LOGGER.error("Aborting.")
        sys.exit(1)
    if choice.strip() == "0":
        return most_recent_backup["identifier"]
    if choice.strip() in [str(i + 1) for i in range(len(revisions))]:
        return revisions[int(choice.strip()) - 1]["identifier"]
    return choice


@click.argument(
    "revisions", type=str, required=False, nargs=-1, metavar="[REVISION]..."
)
@_subcommand_init
def delete(repo_root: Path, revisions: tuple[str, ...]):
    """Delete one or more backups."""
    if not revisions:
        revisions = _prompt_for_revisions_to_delete(repo_root)
    try:
        fastforward.delete_backups(repo_root, *revisions)
        LOGGER.log(
            IMPORTANT,
            'Deleted backups are now marked as "loose."'
            " To delete them immediately, run the command:"
            "\n  git gc --aggressive --prune=now",
        )
    except ValueError as whats_that:
        LOGGER.error(whats_that)
        sys.exit(1)


def _prompt_for_revisions_to_delete(repo_root: Path) -> tuple[str, ...]:
    """Offer a history of revisions to delete. Unlike similar prompts, this
    is a multiselect and requires the user to type in each entry (in order to guard
    against accidental deletions)."""
    LOGGER.log(IMPORTANT, "Here is a list of recent GSB-created backups:")
    revisions = history_.show_history(
        repo_root, tagged_only=False, since_last_tagged_backup=True, numbering=None
    )
    revisions.extend(history_.show_history(repo_root, limit=3, numbering=None))
    if len(revisions) == 0:
        LOGGER.warning("No GSB revisions found. Trying Git.")
        revisions = history_.show_history(
            repo_root, limit=10, tagged_only=False, include_non_gsb=True, numbering=None
        )
    LOGGER.log(
        IMPORTANT,
        "\nSelect a backup to delete by identifier, or multiple separated by commas."
        "\nAlternatively, [q]uit and call gsb history yourself to get more revisions.",
    )
    if len(revisions) == 0:
        LOGGER.error("No revisions found!")
        sys.exit(1)

    choices: str = click.prompt(
        "Select a revision or revisions", default="q", show_default=True, type=str
    ).lower()

    if choices.lower().strip() == "q":
        LOGGER.error("Aborting.")
        sys.exit(1)
    return tuple(choice.strip() for choice in choices.strip().split(","))


@click.option(
    "-J",
    "xz_flag",
    is_flag=True,
    flag_value="tar.xz",
    help="Export as a .tar.xz archive.",
)
@click.option(
    "-j",
    "bz2_flag",
    is_flag=True,
    flag_value="tar.bz2",
    help="Export as a .tar.bz2 archive.",
)
@click.option(
    "-z",
    "gz_flag",
    is_flag=True,
    flag_value="tar.gz",
    help="Export as a .tar.gz archive.",
)
@click.option(
    "-t",
    "tar_flag",
    is_flag=True,
    flag_value="tar",
    help="Export as an uncompressed .tar archive.",
)
@click.option(
    "-p",
    "zip_flag",
    is_flag=True,
    flag_value="zip",
    help="Export as a .zip archive.",
)
@click.option(
    "--format",
    "archive_format",
    type=str,
    required=False,
    help=(
        "Format for the archived backup. If not specified,"
        " an appropriate one will be chosen based on your OS."
    ),
    metavar="FORMAT",
)
@click.option(
    "--output",
    "-o",
    type=Path,
    required=False,
    help=(
        "Explicitly specify a filename for the archived backup."
        " The format of the archive will be inferred from the extension"
        " unless a format flag is provided."
    ),
    metavar="FILENAME",
)
@click.argument(
    "revision",
    type=str,
    required=False,
)
@_subcommand_init
def export(
    repo_root: Path,
    revision: str | None,
    output: Path | None,
    **format_flags,
):
    """Create a stand-alone archive specified REVISION."""
    print(format_flags)
    specified_formats: list[str] = [value for value in format_flags.values() if value]

    if len(specified_formats) > 1:
        LOGGER.error("Conflicting values given for archive format")
        sys.exit(1)

    if revision is None:
        revision = _prompt_for_a_recent_revision(repo_root)

    if len(specified_formats) == 1:
        archive_format = specified_formats[0]
        if archive_format.startswith("."):
            archive_format = archive_format[1:]
        if output is None:
            output = Path(
                export_.generate_archive_name(
                    Manifest.of(repo_root).name, revision, extension=archive_format
                )
            )
        else:
            output = output.parent / (output.name + f".{archive_format}")

    try:
        export_.export_backup(repo_root, revision, output)
    except ValueError as whats_that:
        LOGGER.error(whats_that)
        sys.exit(1)


@click.argument(
    "pytest_args",
    nargs=-1,
)
@gsb.command(context_settings={"ignore_unknown_options": True})
def test(pytest_args: tuple[str, ...]):  # pragma: no cover
    """Run the GSB test suite to ensure that it is running correctly on your system.
    Requires you to have installed GSB with the test extra
    (i.e. pipx install gsb[test])."""
    import pytest

    pytest.main(["--pyargs", "gsb.test", *pytest_args])



"""Tests for restoring backups"""
import subprocess
import time

import pytest

from gsb import _git, backup, onboard, rewind
from gsb.history import get_history
from gsb.manifest import MANIFEST_NAME


@pytest.fixture
def repo(tmp_path, patch_tag_naming):
    my_game_data = tmp_path / "best game ever"
    my_save_data = my_game_data / "save" / "data.txt"
    my_save_data.parent.mkdir(parents=True)
    my_save_data.touch()
    _git.init(my_game_data)
    _git.add(my_game_data, my_save_data.name)
    _git.commit(
        my_game_data,
        "Back when the world was new",
        _committer=("Old Man", "old@man"),
    )
    onboard.create_repo(my_game_data, "save")

    for i in range(10):
        my_save_data.write_text(f"{i}\n")
        if i % 2 == 0:
            (my_save_data.parent / ".boop").touch()
            (my_save_data.parent / ".beep").unlink(missing_ok=True)
        else:
            (my_save_data.parent / ".boop").unlink()
            (my_save_data.parent / ".beep").touch()
        backup.create_backup(my_game_data, None)
        if i % 3 == 0:
            backup.create_backup(my_game_data, "Checkpoint")
    my_save_data.write_text("Sneaky sneaky\n")
    yield my_game_data


class TestRestoreBackup:
    @pytest.mark.parametrize("root_type", ("no_folder", "no_git", "no_manifest"))
    def test_raises_when_theres_no_gsb_repo(self, tmp_path, root_type):
        random_folder = tmp_path / "random folder"
        reference = "blah"
        if root_type != "no_folder":
            random_folder.mkdir()
        if root_type == "no_manifest":
            _git.init(random_folder)
            (random_folder / ".something").touch()
            _git.add(random_folder, ".something")
            commit = _git.commit(random_folder, "placeholder")
            reference = commit.hash
        with pytest.raises(OSError):
            rewind.restore_backup(random_folder, reference)

    def test_raises_when_revision_is_invalid(self, repo):
        with pytest.raises(ValueError):
            rewind.restore_backup(repo, "not-a-thing")

    def test_restores_file_content_to_a_previous_tag(self, repo):
        rewind.restore_backup(repo, "gsb2023.07.13")
        assert (repo / "save" / "data.txt").read_text() == "6\n"

    def test_restore_tag_references_the_restored_version(self, repo):
        rewind.restore_backup(repo, "gsb2023.07.10")
        assert (
            get_history(repo, limit=1)[0]["identifier"]
            == "gsb2023.07.16.restore_of_gsb2023.07.10"
        )

    def test_restores_file_content_to_a_previous_commit(self, repo):
        commit_two = list(get_history(repo, tagged_only=False))[-4]
        assert not commit_two["identifier"].startswith("gsb")  # not a tag
        rewind.restore_backup(repo, commit_two["identifier"])
        assert (repo / "save" / "data.txt").read_text() == "2\n"

    def test_restores_a_deleted_file(self, repo):
        assert not (repo / "save" / ".boop").exists()
        rewind.restore_backup(repo, "gsb2023.07.13")
        assert (repo / "save" / ".boop").exists()

    def test_deletes_a_new_file(self, repo):
        assert (repo / "save" / ".beep").exists()
        rewind.restore_backup(repo, "gsb2023.07.13")
        assert not (repo / "save" / ".beep").exists()

    def test_unstaged_changes_are_commited_before_restore(self, repo):
        last_backup = get_history(repo, tagged_only=False, limit=1)[0]
        assert (repo / "save" / "data.txt").read_text() == "Sneaky sneaky\n"
        rewind.restore_backup(repo, last_backup["identifier"])
        assert (repo / "save" / "data.txt").read_text() == "9\n"
        all_backups = list(get_history(repo, tagged_only=False))

        assert all_backups[2] == last_backup  # because reverse-chronological

        pre_restore_backup = all_backups[1]

        rewind.restore_backup(repo, pre_restore_backup["identifier"])
        assert (repo / "save" / "data.txt").read_text() == "Sneaky sneaky\n"

    def test_restored_state_gets_backed_up(self, repo):
        rewind.restore_backup(repo, "gsb2023.07.10")

        with pytest.raises(ValueError, match="othing to"):
            _git.add(repo, ["saves"])
            _git.commit(repo, "Test")

        assert (
            "restore"
            in next(iter(get_history(repo, tagged_only=False)))["description"].lower()
        )

    def test_gsb_manifest_and_gitignore_are_not_rewound(self, repo):
        with (repo / ".gitignore").open("a") as ignore:
            ignore.write("nothingtosee\n")
        with (repo / MANIFEST_NAME).open("a") as manifest:
            manifest.write("# it's a comment\n")

        rewind.restore_backup(repo, "gsb2023.07.10")

        with pytest.raises(ValueError, match="othing to"):
            _git.add(repo, ["saves"])
            _git.commit(repo, "Test")

        assert "nothingtosee" in (repo / ".gitignore").read_text().splitlines()
        assert "# it's a comment" in (repo / MANIFEST_NAME).read_text().splitlines()

    def test_gsb_manifest_and_gitignore_can_be_rewound(self, repo):
        with (repo / ".gitignore").open("a") as ignore:
            ignore.write("nothingtosee\n")
        with (repo / MANIFEST_NAME).open("a") as manifest:
            manifest.write("# it's a comment\n")

        rewind.restore_backup(repo, "gsb2023.07.10", keep_gsb_files=False)

        with pytest.raises(ValueError, match="othing to"):
            _git.add(repo, ["saves"])
            _git.commit(repo, "Test")

        assert "nothingtosee" not in (repo / ".gitignore").read_text()
        assert "# it's a comment" not in (repo / MANIFEST_NAME).read_text()

    def test_can_rewind_to_pre_gsb_state(self, repo):
        first_commit = get_history(repo, tagged_only=False, include_non_gsb=True)[-1]

        rewind.restore_backup(repo, first_commit["identifier"])

        assert (repo / MANIFEST_NAME).exists()

    def test_rewind_tag_naming_doesnt_cause_conflicts(self, tmp_path):
        repo_root = tmp_path / "repossess"
        repo_root.mkdir()
        onboard.create_repo(repo_root, "furniture")
        restore_point = get_history(repo_root, limit=1)[0]
        (repo_root / "furniture").mkdir()
        (repo_root / "furniture" / "sofa").write_text("I'm the king!")
        backup.create_backup(repo_root)
        time.sleep(1)  # blergh
        rewind.restore_backup(repo_root, restore_point["identifier"])
        assert not (repo_root / "furniture" / "sofa").exists()


class TestCLI:
    def test_no_args_initiates_prompt_in_cwd(self, repo):
        result = subprocess.run(
            ["gsb", "rewind"], cwd=repo, capture_output=True, input="q\n".encode()
        )

        assert (
            "Select one by number or identifier"
            in result.stderr.decode().strip().splitlines()[-2]
        )

    def test_passing_in_a_custom_root(self, repo):
        result = subprocess.run(
            ["gsb", "rewind", "--path", repo.name],
            cwd=repo.parent,
            capture_output=True,
            input="q\n".encode(),
        )

        assert (
            "Select one by number or identifier"
            in result.stderr.decode().strip().splitlines()[-2]
        )

    def test_restoring_to_tag_by_argument(self, repo):
        _ = subprocess.run(
            ["gsb", "rewind", "gsb2023.07.12"], cwd=repo, capture_output=True
        )

        assert (repo / "save" / "data.txt").read_text() == "3\n"

    @pytest.mark.parametrize("how", ("short", "full"))
    def test_restoring_to_commit(self, repo, how):
        some_commit = list(_git.log(repo))[-5].hash
        if how == "short":
            some_commit = some_commit[:8]

        _ = subprocess.run(
            ["gsb", "rewind", some_commit],
            cwd=repo,
            capture_output=True,
        )

        assert (repo / "save" / "data.txt").read_text() == "2\n"

    @pytest.mark.parametrize("how", ("by_tag", "by_index"))
    def test_restoring_by_prompt(self, repo, how):
        if how == "by_tag":
            choice = "gsb2023.07.13"
        else:
            choice = "2"

        _ = subprocess.run(
            ["gsb", "rewind"],
            cwd=repo,
            capture_output=True,
            input=f"{choice}\n".encode(),
        )

        assert (repo / "save" / "data.txt").read_text() == "6\n"

    @pytest.mark.parametrize("is_gsb", (True, False), ids=("gsb", "non-gsb"))
    def test_most_recent_backup_is_a_choice(self, repo, is_gsb):
        if is_gsb:
            commit_hash = backup.create_backup(repo)
        else:
            _git.add(repo, ("save",))
            commit_hash = _git.commit(
                repo, "Sneakier and Sneakier", _committer=("you-ser", "me@computer")
            ).hash

        (repo / "save" / "data.txt").write_text("Unsaved changes!\n")

        result = subprocess.run(
            ["gsb", "rewind"], cwd=repo, capture_output=True, input="0\n".encode()
        )

        assert f"0. {commit_hash[:8]}" in result.stderr.decode()

        # check that the contents were restored
        assert (repo / "save" / "data.txt").read_text() == "Sneaky sneaky\n"

    @pytest.mark.parametrize("how", ("by_argument", "by_prompt"))
    def test_unknown_revision_raises_error(self, repo, how):
        args = ["gsb", "rewind"]
        answers = [""]
        if how == "by_argument":
            args.append("not_a_thing")
        else:
            answers.insert(0, "not_a_thing")

        result = subprocess.run(
            args,
            cwd=repo,
            capture_output=True,
            input="\n".join(answers).encode(),
        )

        assert result.returncode == 1
        assert "Could not find" in result.stderr.decode().strip().splitlines()[-1]

    def test_running_on_repo_with_no_tags_retrieves_gsb_commits(self, tmp_path):
        """Like, I guess if the user deleted the initial backup"""
        repo = tmp_path / "repo"
        repo.mkdir()
        something = repo / "file"
        something.touch()
        _git.init(repo)
        _git.add(repo, [something.name])
        commit_hash = _git.commit(repo, "Hello").hash[:8]

        result = subprocess.run(
            ["gsb", "rewind"], cwd=repo, capture_output=True, input="q\n".encode()
        )
        assert f"1. {commit_hash}" in result.stderr.decode().strip().splitlines()[1]

    def test_running_on_non_gsb_prompts_with_git_commits(self, tmp_path):
        repo = tmp_path / "repo"
        repo.mkdir()
        something = repo / "file"
        something.touch()
        _git.init(repo)
        _git.add(repo, [something.name])
        commit_hash = _git.commit(repo, "Hello", _committer=("Testy", "Testy")).hash[:8]

        result = subprocess.run(
            ["gsb", "rewind"], cwd=repo, capture_output=True, input="q\n".encode()
        )
        log_lines = result.stderr.decode().strip().splitlines()

        assert "No GSB revisions found" in log_lines[1]
        assert f"1. {commit_hash}" in log_lines[2]

    def test_running_on_empty_repo_raises(self, tmp_path):
        repo = tmp_path / "repo"
        repo.mkdir()
        something = repo / "file"
        something.touch()
        _git.init(repo)

        result = subprocess.run(["gsb", "rewind"], cwd=repo, capture_output=True)
        assert result.returncode == 1
        assert "No revisions found" in result.stderr.decode().strip().splitlines()[-1]

    def test_force_rewind_of_required_files(self, repo):
        with (repo / MANIFEST_NAME).open("a") as manifest:
            manifest.write("# it's a comment\n")

        _ = subprocess.run(
            ["gsb", "rewind", "gsb2023.07.12", "--include_gsb_settings"],
            cwd=repo,
            capture_output=False,
        )

        assert "# it's a comment" not in (repo / MANIFEST_NAME).read_text()



from pathlib import Path

from setuptools import setup

import versioneer

long_description = (Path(__file__).parent / "README.md").read_text()

setup(
    name="gsb",
    python_requires=">=3.11",
    description="Game Save Backups: A Git-Based Tool for Managing Your Save States",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Gili "OpenBagTwo" Barlev',
    url="https://github.com/OpenBagTwo/gsb",
    packages=[
        "gsb",
        "gsb.test",
    ],
    package_data={"gsb": ["py.typed"]},
    entry_points={
        "console_scripts": [
            "gsb = gsb.cli:gsb",
        ]
    },
    license="GPL v3",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    install_requires=["pygit2>=1.12", "click>=8", "pathvalidate>=2.5"],
    extras_require={"test": ["pytest>=7", "pytest-cov>=4"]},
)



"""Configuration definition for an individual GSB-managed save"""
import datetime as dt
import json
import logging
import tomllib
from pathlib import Path
from typing import Any, NamedTuple, Self, TypeAlias

from ._version import get_versions

LOGGER = logging.getLogger(__name__)

MANIFEST_NAME = ".gsb_manifest"

_ManifestDict: TypeAlias = dict[str, str | tuple[str, ...]]


class Manifest(NamedTuple):
    """Save-specific configuration

    Attributes
    ----------
    root : Path
        The directory containing the save / repo
    name : str
        The name / alias of the repo
    patterns : tuple of str
        The glob match-patterns that determine which files get tracked
    """

    root: Path
    name: str
    patterns: tuple[str, ...]

    @classmethod
    def of(cls, repo_root: Path) -> Self:
        """Read the manifest of the specified GSB repo

        Parameters
        ----------
        repo_root : Path
            The root directory of the gsb-managed repo

        Returns
        -------
        Manifest
            the parsed manifest

        Raises
        ------
        ValueError
            If the configuration cannot be parsed
        OSError
            If the file does not exist or cannot otherwise be read
        """
        LOGGER.debug("Loading %s from %s", MANIFEST_NAME, repo_root)
        as_dict: dict[str, Any] = {"root": repo_root}
        contents: _ManifestDict = tomllib.loads((repo_root / MANIFEST_NAME).read_text())
        for key, value in contents.items():
            if key in Manifest._fields:
                if isinstance(value, list):
                    value = tuple(value)
                as_dict[key] = value
        if "name" not in as_dict:
            as_dict["name"] = repo_root.resolve().name
        return cls(**as_dict)

    def write(self) -> None:
        """Write the manifest to file, overwriting any existing configuration

        Returns
        -------
        None

        Notes
        -----
        The location and name of this file is controlled by the `root` attribute
        and the `MANIFEST_NAME` constant, respectively, and cannot be overridden

        Raises
        ------
        OSError
            If the destination folder (`root`) does not exist or cannot be
            written to
        """
        as_dict = {
            "generated_by_gsb": get_versions()["version"],
            "last_modified": dt.datetime.now().isoformat(sep=" "),
        }
        for attribute, value in self._asdict().items():  # pylint: disable=no-member
            #                  see: https://github.com/pylint-dev/pylint/issues/7891
            if attribute == "root":
                continue
            as_dict[attribute] = value

        as_toml = _to_toml(as_dict)

        LOGGER.debug("Writing %s to %s", MANIFEST_NAME, self.root)
        (self.root / MANIFEST_NAME).write_text(as_toml)


def _to_toml(manifest: _ManifestDict) -> str:
    """While Python 3.11 added native support for *parsing* TOML configurations,
    it didn't include an API for *writing* them (this was an intentional part
    of the PEP:
    https://peps.python.org/pep-0680/#including-an-api-for-writing-toml).

    Because the Manifest class is so simple, I'm rolling my own writer rather
    than adding a dependency on a third-party library. That being said, I'm
    abstracting that writer out in case I change my mind later. :D

    Parameters
    ----------
    manifest : dict
        A dict version of the manifest containing the entries that should be
        written to file

    Returns
    -------
    str
        The manifest serialized as a TOML-compatible str

    Notes
    -----
    This doesn't take an actual Manifest as a parameter so we can choose to
    omit some attributes (`root`) and add others (versioning metadata)
    """
    dumped = ""
    for key, value in manifest.items():
        dumped += f"{key} = "
        if isinstance(value, str):
            # it's honestly shocking how often I rely on json.dump for str escaping
            dumped += f"{json.dumps(value)}\n"
        else:
            dumped += "["
            for entry in sorted(set(value)):
                dumped += f"\n    {json.dumps(entry)},"
            dumped += "\n]\n"
    return dumped



"""Functionality for restoring to an old backup"""
import logging
from pathlib import Path

from . import _git, backup
from .logging import IMPORTANT

LOGGER = logging.getLogger(__name__)


def generate_restore_tag_name(revision: str) -> str:
    """Generate a new calver-ish tag name

    Parameters
    ----------
    revision : str
        The commit hash or tag name of the backup to restore

    Returns
    -------
    str
        A tag name that indicates both the time a backup was restored and the
        identifier of the original revision
    """
    return f"{backup.generate_tag_name()}.restore_of_{revision}"


def restore_backup(repo_root: Path, revision: str, keep_gsb_files: bool = True) -> str:
    """Rewind to a previous backup state and create a new backup

    Parameters
    ----------
    repo_root : Path
        The directory containing the GSB-managed repo
    revision : str
        The commit hash or tag name of the backup to restore
    keep_gsb_files : bool, optional
        By default, `.gsb_manifest` and `.gitignore` *will not* be restored
        (that is, the latest versions will be kept). To override this behavior,
        pass in `keep_gsb_files = False`.

    Returns
    -------
    str
        The tag name of the new restored backup

    Notes
    -----
    Before creating the backup, any un-backed up changes will first be backed up

    Raises
    ------
    OSError
        If the specified repo does not exist or is not a GSB-managed repo
    ValueError
        If the specified revision does not exist
    """
    _git.show(repo_root, revision)

    LOGGER.log(
        IMPORTANT, "Backing up any unsaved changes before rewinding to %s", revision
    )
    try:
        orig_head = backup.create_backup(
            repo_root, f"Backing up state before rewinding to {revision}"
        )
    except ValueError as already_backed_up:
        LOGGER.warning("Nothing to back up: %s", already_backed_up)
        orig_head = _git.show(repo_root, "HEAD").hash  # type: ignore[union-attr]

    _git.reset(repo_root, revision, hard=True)
    _git.reset(repo_root, orig_head, hard=False)
    if keep_gsb_files:
        _git.checkout_files(repo_root, orig_head, backup.REQUIRED_FILES)
    return backup.create_backup(
        repo_root,
        f"Restored to {revision}",
        tag_name=generate_restore_tag_name(revision),
    )



"""Functionality for creating standalone backups"""
import os
from pathlib import Path

import pathvalidate

from . import _git
from .manifest import Manifest


def generate_archive_name(
    repo_name: str, revision: str, extension: str | None = None
) -> str:
    """Programmatically generate a name for an archived backup

    Parameters
    ----------
    repo_name : str
        The alias assigned to the GSB-managed repo
    revision : str
        The commit hash or tag name of the backup that's being archived
    extension : str, optional
        The file extension for the archive (thus specifying the archive's format).
        If None is provided, an appropriate one will be chosen based on the
        operating system.

    Returns
    -------
    str
        A (hopefully) descriptive archive filename, including a format-specifying
        extension

    Notes
    -----
    The default choice of extension (and thus format) is:

    - zip for Windows
    - tar.gz for all other systems
    """
    if extension is None:
        extension = "zip" if os.name == "nt" else "tar.gz"
    return pathvalidate.sanitize_filename(f"{repo_name}_{revision}.{extension}")


def export_backup(
    repo_root: Path,
    revision: str,
    archive_path: Path | None = None,
) -> None:
    """Export a backup to a stand-alone archive

    Parameters
    ----------
    repo_root : Path
        The directory containing the GSB-managed repo
    revision : str
        The commit hash or tag name of the backup to archive
    archive_path : Path, optional
        The full path to save the archive, including the filename and the
        extension. If None is provided, one will be automatically generated
        in the current working directory based on the repo's name and the
        specified revision.

    Raises
    ------
    OSError
        If the specified repo does not exist or is not a GSB-managed repo
    ValueError
        If the specified revision does not exist or if the given `archive_path`
        does not have a valid extension
    NotImplementedError
        If the compression schema implied by the `archive_path`'s extension is not
        supported
    """
    if archive_path is None:
        archive_path = Path(
            generate_archive_name(Manifest.of(repo_root).name, revision)
        )

    _git.archive(repo_root, archive_path, revision)
